'''
Created on 5 oct. 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import os
import rasterio
from numpy.random import default_rng
from os import remove
from os.path import exists
from tempfile import TemporaryDirectory
from t4gpd.commons.GeoProcess import GeoProcess


class AbstractRasterGeoProcess(GeoProcess):
    '''
    classdocs
    '''
    @staticmethod
    def _write_and_load(array, metadata, debug, indexes=1):
        rng = default_rng()

        if debug:
            ofile = f"__temporary__{rng.integers(1e12)}.tif"
            if exists(ofile):
                remove(ofile)
            with rasterio.open(ofile, "w", **metadata) as dst:
                if indexes is None:
                    dst.write(array)
                else:
                    dst.write(array, indexes=1)
            result = rasterio.open(ofile)
        else:
            with TemporaryDirectory() as tmpdir:
                ofile = f"{tmpdir}/temporary.tif"
                with rasterio.open(ofile, "w", **metadata) as dst:
                    if indexes is None:
                        dst.write(array)
                    else:
                        dst.write(array, indexes=indexes)
                result = rasterio.open(ofile)
        return result



'''
Created on 23 juin 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import log, power


class ColorTemperature(object):
    '''
    classdocs
    '''

    @staticmethod
    def convert_K_to_RGB(colour_temperature):
        """
        Converts from K to RGB, algorithm courtesy of 
        http://www.tannerhelland.com/4435/convert-temperature-rgb-algorithm-code/
        https://gist.github.com/petrklus/b1f427accdf7438606a6
        """
        def setMinMax(value, minValue, maxValue): return min(
            max(value, minValue), maxValue)

        # range check
        colour_temperature = setMinMax(colour_temperature, 1000, 40000)
        tmp_internal = colour_temperature / 100.0

        # red
        if tmp_internal <= 66:
            red = 255
        else:
            red = 329.698727446 * power(tmp_internal - 60, -0.1332047592)
            red = setMinMax(red, 0, 255)

        # green
        if tmp_internal <= 66:
            green = 99.4708025861 * log(tmp_internal) - 161.1195681661
        else:
            green = 288.1221695283 * power(tmp_internal - 60, -0.0755148492)
        green = setMinMax(green, 0, 255)

        # blue
        if tmp_internal >= 66:
            blue = 255
        elif tmp_internal <= 19:
            blue = 0
        else:
            blue = 138.5177312231 * \
                log(tmp_internal - 10) - 305.0447927307
            blue = setMinMax(blue, 0, 255)

        # return red, green, blue
        return int(round(red)), int(round(green)), int(round(blue))



'''
Created on 26 sep. 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import meshio
from geopandas import GeoDataFrame
from shapely import Polygon
from t4gpd.commons.GeoProcess import GeoProcess


class MedReader(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, inputFile, crs="EPSG:2154"):
        '''
        Constructor
        '''
        self.inputFile = inputFile
        self.crs = crs

    def __analyze(self, mesh):
        dictOfMaterials = mesh.cell_tags
        points = mesh.points
        triangles = mesh.cells_dict["triangle"]
        ntriangles = len(triangles)
        for materials in mesh.cell_data["cell_tags"]:
            if ntriangles == len(materials):
                break
        return dictOfMaterials, points, triangles, materials

    def run(self):
        mesh = meshio.read(self.inputFile)
        dictOfMaterials, points, triangles, materials = self.__analyze(mesh)
        return GeoDataFrame([
            {
                "geometry": Polygon([points[t[0]], points[t[1]], points[t[2]], points[t[0]]]),
                "material": dictOfMaterials[m][0]
            } for t, m in zip(triangles, materials)
        ], crs=self.crs)



'''
Created on 22 aug. 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import rasterio
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class RTLoad(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, filename):
        '''
        Constructor
        '''
        if not isinstance(filename, str):
            raise IllegalArgumentTypeException(
                filename, "input filename (str)")
        self.filename = filename

    def run(self):
        return rasterio.open(self.filename)


"""
r = RTLoad("/home/tleduc/data/mnt_asc/mnt_nantes_2005_1m/MNT_L93_0354_6690.asc").run()
print(type(r))
"""



'''
Created on 11 oct. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class GpkgWriter(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, mapOfGdf, gpkgOutputFile):
        '''
        Constructor
        '''
        if not isinstance(mapOfGdf, dict):
            raise IllegalArgumentTypeException(mapOfGdf, '{ "layername": GeoDataFrame, ... }')
        for layername, gdf in mapOfGdf.items():
            if not isinstance(layername, str):
                raise IllegalArgumentTypeException(layername, 'layername')
            if not isinstance(gdf, GeoDataFrame):
                raise IllegalArgumentTypeException(gdf, 'GeoDataFrame')
        self.mapOfGdf = mapOfGdf

        self.gpkgOutputFile = gpkgOutputFile

    def run(self):
        for layername, df in self.mapOfGdf.items():
            df.to_file(self.gpkgOutputFile, driver='GPKG', layer=layername)

        print(f'{self.gpkgOutputFile} has been written!')
        return None



'''
Created on 14 oct. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import datetime, timezone
from locale import LC_ALL, setlocale
from os import mkdir
from os.path import isdir

from geopandas import clip, GeoDataFrame
from matplotlib_scalebar.scalebar import ScaleBar
from numpy import isnan
from pandas import DataFrame
from shapely.geometry import box
from t4gpd.comfort.indices.PET import PET
from t4gpd.comfort.indices.UTCI import UTCI
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.misc.STCompass import STCompass
from t4gpd.morph.STSquaredBBox import STSquaredBBox
from t4gpd.sun.STHardShadow import STHardShadow
from t4gpd.sun.STTreeHardShadow import STTreeHardShadow
from t4gpd.sun.STTreeHardShadow2 import STTreeHardShadow2

import matplotlib.pyplot as plt
from t4gpd.commons.DataFrameLib import DataFrameLib


class MetrologicalCampaignPlottings(GeoProcess):
    '''
    classdocs
    '''
    COLORS = ['black', 'blue', 'green']

    def __init__(self, tracks, waypoints, dfImuMob, dfStat1=None, dfStat2=None,
                 dfStat3=None, buildings=None, sidewalks=None, trees=None,
                 dstDir='.', ext='pdf'):
        '''
        Constructor
        '''
        if not isinstance(tracks, GeoDataFrame):
            raise IllegalArgumentTypeException(tracks, 'GeoDataFrame')
        self.tracks = tracks

        if not isinstance(waypoints, GeoDataFrame):
            raise IllegalArgumentTypeException(waypoints, 'GeoDataFrame')
        self.waypoints = waypoints

        if not isinstance(dfImuMob, GeoDataFrame):
            raise IllegalArgumentTypeException(dfImuMob, 'GeoDataFrame')
        self.dfImuMob = dfImuMob

        if not (dfStat1 is None or isinstance(dfStat1, DataFrame)):
            raise IllegalArgumentTypeException(dfStat1, 'DataFrame or None')
        self.dfStat1 = dfStat1

        if not (dfStat2 is None or isinstance(dfStat2, DataFrame)):
            raise IllegalArgumentTypeException(dfStat2, 'DataFrame or None')
        self.dfStat2 = dfStat2

        if not (dfStat3 is None or isinstance(dfStat3, DataFrame)):
            raise IllegalArgumentTypeException(dfStat3, 'DataFrame or None')
        self.dfStat3 = dfStat3

        if not (buildings is None or isinstance(buildings, GeoDataFrame)):
            raise IllegalArgumentTypeException(buildings, 'GeoDataFrame or None')
        self.buildings = buildings

        if not (sidewalks is None or isinstance(sidewalks, GeoDataFrame)):
            raise IllegalArgumentTypeException(sidewalks, 'GeoDataFrame or None')
        self.sidewalks = sidewalks

        if not (trees is None or isinstance(trees, GeoDataFrame)):
            raise IllegalArgumentTypeException(trees, 'GeoDataFrame or None')
        self.trees = trees

        if not isdir(dstDir):
            mkdir(dstDir)
        self.dstDir = dstDir

        if not (ext in ['pdf', 'png']):
            raise IllegalArgumentTypeException(ext, '"pdf" or "png"')
        self.ext = ext

    def __graphicBoard(self, imuMob):
        trackId = int(imuMob.track.tolist()[0])
        tracks = self.tracks[self.tracks.id == trackId].copy(deep=True)
        waypoints = self.waypoints[(self.waypoints.id // 100) == trackId].copy(deep=True)

        t0, t1 = imuMob.timestamp.min(), imuMob.timestamp.max()

        nrows, ncols, magn = 3, 3, 0.5
        fig, axes = plt.subplots(nrows, ncols, squeeze=False,
                                 figsize=(ncols * magn * 8.26, nrows * magn * 8.26))
        setlocale(LC_ALL, 'en_US.UTF8')
        fig.suptitle(f'''Track #{trackId} --- {t0.strftime('%b. %d %Y')}
From {t0.strftime('%X')} to {t1.strftime('%X')}''', size=24)
        # -----
        roi = self.__map(axes, 0, 0, t0, t1, tracks, waypoints)
        self.__subplot(axes, 0, 1, waypoints, imuMob,
                       ['AirTC_Avg', 'Temp_C_Avg(1)', 'Temp_C_Avg(2)'],
                       '$^\circ$C')
        self.__subplot(axes, 0, 2, waypoints, imuMob, ['RH_Avg'], '%', minmax=[15, 85])

        self.__subplot2(axes, 1, 0, roi, t0, t1, self.dfStat1, self.dfStat2, self.dfStat3)
        self.__subplot(axes, 1, 1, waypoints, imuMob, ['SR01Up_3_Avg', 'IR01DnCo_3_Avg'],
                       'W.m$^{_2}$')
        self.__subplot(axes, 1, 2, waypoints, imuMob, ['WS_ms_Avg'], 'm.s$^{-1}$')

        self.__subplot(axes, 2, 0, waypoints, imuMob, ['T_mrt'], '$^\circ$C')
        self.__subplot(axes, 2, 1, waypoints, imuMob, ['PET'], '$^\circ$C',
                       thermalPerceptionRanges=PET.thermalPerceptionRanges)
        self.__subplot(axes, 2, 2, waypoints, imuMob, ['UTCI'], '$^\circ$C',
                       thermalPerceptionRanges=UTCI.thermalPerceptionRanges)
        # -----
        plt.savefig(f'{self.dstDir}/{trackId}_{t0.strftime("%Y%m%d_%H%M")}.{self.ext}', bbox_inches='tight')
        # plt.show()
        plt.close(fig)

    @staticmethod
    def __shadows_1(t0, t1, buildings):
        t0 = datetime(*t0.timetuple()[:6], tzinfo=timezone.utc)
        sob = STHardShadow(buildings, t0, occludersElevationFieldname='HAUTEUR',
                           altitudeOfShadowPlane=0, aggregate=True,
                           tz=timezone.utc, model='pysolar').run()
        return sob

    @staticmethod
    def __shadows_2(t0, t1, trees):
        t0 = datetime(*t0.timetuple()[:6], tzinfo=timezone.utc)
        # sot = STTreeHardShadow(
        #     trees, t0, treeHeightFieldname='h_arbre',
        #     treeCrownRadiusFieldname='r_houppier', altitudeOfShadowPlane=0.0,
        #     aggregate=True, tz=timezone.utc, model='pysolar', npoints=32).run() 
        sot = STTreeHardShadow2(trees, t0, treeHeightFieldname='h_arbre',
            treeCrownHeightFieldname='h_houppier',
            treeUpperCrownRadiusFieldname='up_rad',
            treeLowerCrownRadiusFieldname='down_rad',
            altitudeOfShadowPlane=0.0, aggregate=False, tz=timezone.utc,
            model='pysolar', npoints=32).run() 
        return sot

    def __map(self, axes, nr, nc, t0, t1, tracks, waypoints):
        ax = axes[nr, nc]

        roi = STSquaredBBox(tracks, buffDist=10.0).run()
        minx, miny, maxx, maxy = roi.total_bounds

        if not self.buildings is None:
            # SHADOWS
            buildings = clip(self.buildings, roi)
            sob = MetrologicalCampaignPlottings.__shadows_1(t0, t1, buildings)
            sob.plot(ax=ax, color='lightgrey', linewidth=0, alpha=0.5)

            self.buildings.plot(ax=ax, color='darkgrey', linewidth=0)

        if not self.trees is None:
            # SHADOWS
            trees = clip(self.trees, roi)
            if (0 < len(trees)):
                sot = MetrologicalCampaignPlottings.__shadows_2(t0, t1, trees)
                sot.plot(ax=ax, color='lightgreen', linewidth=0, alpha=0.5)

                self.trees.plot(ax=ax, color='green', marker='.')
            
        if not self.sidewalks is None:
            self.sidewalks.plot(ax=ax, color='lightgrey', linewidth=0.4)

        tracks.plot(ax=ax, color='black', linewidth=0.8)
        waypoints.plot(ax=ax, color='black', marker='+')
        waypoints[waypoints.id % 4 == 0].apply(lambda x: ax.annotate(
            text=x.id, xy=x.geometry.coords[0],
            color='blue', size=10, ha='center'), axis=1);

        scalebar = ScaleBar(dx=1.0, units='m', length_fraction=None, width_fraction=0.015,
                            location='lower left', frameon=True)
        ax.add_artist(scalebar)
        STCompass(ax, roi.total_bounds, tracks.crs).run()
        ax.axis('off')
        ax.axis([minx, maxx, miny, maxy])
        return box(*roi.total_bounds)

    def __subplot(self, axes, nr, nc, waypoints, df, indics, unit,
                  minmax=None, thermalPerceptionRanges=None):
        indics = [indic for indic in indics if indic in df]

        ax = axes[nr, nc]

        if minmax is None:
            # ymin = min([df[indic].min() for indic in indics])
            # ymax = max([df[indic].max() for indic in indics])
            ymin = min([self.dfImuMob[indic].min() for indic in indics])
            ymax = max([self.dfImuMob[indic].max() for indic in indics])
        else:
            ymin, ymax = minmax
        deltaY = ymax - ymin

        if not isnan(deltaY):
            _ymin, _ymax = ymin, ymax
            for i, indic in enumerate(indics):
                ax.plot(100 * df.curv_absc, df[indic], color=self.COLORS[i],
                        label=f'{indic}')

                if not thermalPerceptionRanges is None:
                    tpr = thermalPerceptionRanges()
                    _ymin = min(ymin, tpr['Comfortable']['max'])
                    _ymax = max(ymax, tpr['Comfortable']['min'])
                    for lbl, v in tpr.items():
                        if (_ymin <= v['min'] <= _ymax) or (_ymin <= v['max'] <= _ymax):
                            vMin, vMax = max(v['min'], _ymin), min(v['max'], _ymax)
                            ax.axhspan(vMin, vMax, alpha=0.25, color=v['color'],
                                       linewidth=0, label=lbl)

            ax.vlines(100 * waypoints.curv_absc, _ymin, _ymax, linestyles='dotted',
                      linewidth=0.1, color='grey')
            for gid, x in zip(waypoints.id, 100 * waypoints.curv_absc):
                ax.text(x, _ymax + 0.02 * deltaY, gid, ha='center', va='bottom',
                        rotation='vertical', color='grey', size=8)
            ax.set_ylim(_ymin - 0.2 * deltaY, _ymax + 0.2 * deltaY)

            if (2 == nr):
                ax.set_xlabel('Curvilinear Abscissa [%]', size=8)
            ax.set_ylabel(f'[{unit}]', size=8)
            ax.legend(loc='lower center', ncols=2, prop={'size': 8})

            return False
        return True

    def __subplot2(self, axes, nr, nc, roi, t0, t1, dfStat1, dfStat2, dfStat3):
        ax = axes[nr, nc]
        ymin, ymax = float('inf'), -float('inf')

        if dfStat1 is not None:
            ymin, ymax = min(ymin, dfStat1.Tair.min()), max(ymax, dfStat1.Tair.max())
            df1 = dfStat1[ (t0 <= dfStat1.timestamp) & (dfStat1.timestamp <= t1)].copy(deep=True)
            for label in df1.sensorFamily.unique():
                _df1 = df1[ df1.sensorFamily == label]
                X = _df1.timestamp.apply(lambda t: (t - t0).seconds)
                ax.plot(X, _df1.Tair, label=label, linestyle='-.', color='black')

        if dfStat2 is not None:
            ymin, ymax = min(ymin, dfStat2.Tair.min()), max(ymax, dfStat2.Tair.max())
            df2 = dfStat2[ (t0 <= dfStat2.timestamp) & (dfStat2.timestamp <= t1)].copy(deep=True)
            df2 = df2[ df2.geometry.apply(lambda g: g.within(roi)) ].copy(deep=True)
            for label2a in df2.sensorFamily.unique():
                for label2b in df2.sensorId.unique():
                    _df2 = df2[ (df2.sensorFamily == label2a) & (df2.sensorId == label2b)]
                    X = _df2.timestamp.apply(lambda t: (t - t0).seconds)
                    ax.plot(X, _df2.Tair, label=f'{label2a}-{label2b}', linestyle='-', color='red')

        if dfStat3 is not None:
            df3 = dfStat3[ dfStat3.timestamp.apply(lambda dt: dt.date() == t0.date()) ]
            ymin, ymax = min(ymin, df3.Tair.min()), max(ymax, df3.Tair.max())
            for label in df3.sensorFamily.unique():
                _df3 = df3[ df3.sensorFamily == label ]
                T0 = DataFrameLib.interpolate(_df3, 'timestamp', 'Tair', t0)
                T1 = DataFrameLib.interpolate(_df3, 'timestamp', 'Tair', t1)
                ax.plot([0, (t1 - t0).seconds], [T0, T1], label=label, linestyle=':')

        deltaY = ymax - ymin
        ax.set_ylim(ymin - 0.2 * deltaY, ymax + 0.2 * deltaY)

        ax.set_xlabel('Timestamp [s]', size=8)
        ax.set_ylabel('[$^\circ$C]', size=8)
        ax.legend(loc='lower center', ncols=2, prop={'size': 8})

    def run(self):
        subtracks = self.dfImuMob.subtrack.unique().tolist()
        for subtrack in subtracks:
            df = self.dfImuMob[self.dfImuMob.subtrack == subtrack].copy(deep=True)
            self.__graphicBoard(df)
            # break

'''
from datetime import datetime, timezone
from geopandas import read_file
from os import path

from t4gpd.comfort.indices.TmrtRadiometer import TmrtRadiometer
from t4gpd.morph.geoProcesses.STGeoProcess import STGeoProcess
from t4gpd.picoclim.JoinByTimeDistance import JoinByTimeDistance
from t4gpd.picoclim.MetrologicalCampaignReader import MetrologicalCampaignReader
from t4gpd.picoclim.SnapImuOnTrackUsingWaypoints import SnapImuOnTrackUsingWaypoints
from t4gpd.sun.STHardShadow import STHardShadow

#~ ======================================================================
if path.exists('c:/Users/tleduc'):
    HOMEDIR = 'c:/Users/tleduc'
elif path.exists('/home/tleduc'):
    HOMEDIR = '/home/tleduc'

BDTDIR = f'{HOMEDIR}/data/bdtopo/BDTOPO_3-0_TOUSTHEMES_SHP_LAMB93_D044_2021-12-15/\
BDTOPO/1_DONNEES_LIVRAISON_2021-12-00166/BDT_3-0_SHP_LAMB93_D044-ED2021-12-15'

dirName1 = '/home/tleduc/prj/nm-ilots-frais/terrain/220711'
# dirName1 = '/home/tleduc/prj/nm-ilots-frais/terrain/220713'
dirName2 = '/home/tleduc/prj/nm-ilots-frais/information'

#~ -----
static, tracks, waypoints, dfImu, dfMob, dfStat1, dfStat2, dfMeteoFr = \
    MetrologicalCampaignReader(dirName1, dirName2).run()
dfImu = dfImu.to_crs(tracks.crs)
imu = SnapImuOnTrackUsingWaypoints(dfImu, tracks, waypoints).run()
dfImuMob = JoinByTimeDistance(imu, dfMob, left_on='timestamp', right_on='timestamp').run()

op1 = TmrtRadiometer(dfImuMob)
dfImuMob1 = STGeoProcess([op1], dfImuMob).run()

op2 = PET(dfImuMob1, 'AirTC_Avg', 'RH_Avg', 'WS_ms_Avg', 'T_mrt')
op3 = UTCI(dfImuMob1, 'AirTC_Avg', 'RH_Avg', 'WS_ms_Avg', 'T_mrt')
dfImuMob123 = STGeoProcess([op2, op3], dfImuMob1).run()

#~ -----
bigroi = STSquaredBBox(tracks, buffDist=100.0).run()
buildings = read_file(f'{BDTDIR}/BATI/BATIMENT.shp', mask=bigroi)
sidewalks = read_file(f'{HOMEDIR}/data/nantes_metropole/VOIE_LIM_CHAUSSEE_L.shp', mask=bigroi)

#~ -----
# t0 = datetime(2022, 7, 13, 10, 43, 34)
# sob = STHardShadow(buildings, [t0], occludersElevationFieldname='HAUTEUR',
#                    altitudeOfShadowPlane=0, aggregate=True, tz=timezone.utc, model='pysolar').run()

MetrologicalCampaignPlottings(tracks, waypoints, dfImuMob123, dfStat1, dfStat2,
                              dfMeteoFr, buildings, sidewalks, ext='png').run()
'''



'''
Created on 15 sept. 2022

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import warnings
from zoneinfo import ZoneInfo

from pandas import read_csv
from t4gpd.commons.WarnUtils import WarnUtils
from t4gpd.io.AbstractReader import AbstractReader


class CampbellSciReader(AbstractReader):
    '''
    classdocs

    https://www.campbellsci.com/cr1000x
    '''
    EXPECTED_FIELDS = ["timestamp", "AirTC_Avg", "RH_Avg", "WindDir", "WS_ms_Avg",
                       "SR01Up_1_Avg", "SR01Dn_1_Avg", "Albedo_1_Avg", "IR01UpCo_1_Avg",
                       "IR01DnCo_1_Avg", "SR01Up_2_Avg", "SR01Dn_2_Avg", "Albedo_2_Avg",
                       "IR01UpCo_2_Avg", "IR01DnCo_2_Avg", "SR01Up_3_Avg", "SR01Dn_3_Avg",
                       "Albedo_3_Avg", "IR01UpCo_3_Avg", "IR01DnCo_3_Avg", "Temp_C_Avg(1)",
                       "Temp_C_Avg(2)", "sensorFamily"]

    def __init__(self, inputFile, tzinfo="Europe/Paris"):
        '''
        Constructor
        '''
        warnings.formatwarning = WarnUtils.format_Warning_alt
        self.inputFile = inputFile
        self.tzinfo = ZoneInfo(tzinfo)

    def __getVersion(self):
        V1 = '"TOA5","CR1000XSeries","CR1000X","11356","CR1000X.Std.03.02","CPU:Programme_CS_V2_2sec.CR1X"'
        V2 = '"TOA5","CR1000XSeries","CR1000X","11356","CR1000X.Std.03.02","CPU:Programme_CS_2sec_thermocouple.CR1X"'
        with CampbellSciReader.opener(self.inputFile) as f:
            for line in f.readlines():
                if line.startswith(V1):
                    return 1
                elif line.startswith(V2):
                    return 2
                raise NotImplementedError(f'\n\n\t*** CampbellSciReader:: unknown file format version! ***')

    def __readV1(self):
        # df = read_csv(self.inputFile, header=None, names=self.FIELD_NAMES,
        #               na_values="NAN", parse_dates=[0], sep=",", skiprows=4)
        USECOLS = [0, 9, 12, 15, 16, 32, 33, 40, 44, 45, 46, 47, 54, 58,
                   59, 60, 61, 68, 72, 73]
        df = read_csv(CampbellSciReader.opener(self.inputFile), header=0,
                      na_values="NAN", parse_dates=[0], sep=",",
                      skiprows=[0, 2, 3], usecols=USECOLS)

        dfUnits = read_csv(CampbellSciReader.opener(self.inputFile), header=0,
                           nrows=1, sep=",", skiprows=[0], usecols=USECOLS)
        if ("meters/second" != dfUnits["WS_ms_Avg"].squeeze().lower()):
            warnings.warn(f'"WS_ms_Avg" is expected to be in m/s!')

        df["Temp_C_Avg(1)"] = None
        df["Temp_C_Avg(2)"] = None
        return df

    def __readV2(self):
        # df = read_csv(self.inputFile, header=None, names=self.FIELD_NAMES,
        #               na_values="NAN", parse_dates=[0], sep=",", skiprows=4)
        USECOLS = [0, 9, 12, 15, 16, 32, 33, 40, 44, 45, 46, 47, 54, 58,
                   59, 60, 61, 68, 72, 73, 74, 79]
        df = read_csv(CampbellSciReader.opener(self.inputFile), header=0,
                      na_values="NAN", parse_dates=[0], sep=",",
                      skiprows=[0, 2, 3], usecols=USECOLS)

        dfUnits = read_csv(CampbellSciReader.opener(self.inputFile), header=0,
                           nrows=1, sep=",", skiprows=[0], usecols=USECOLS)
        if ("meters/second" != dfUnits["WS_ms_Avg"].squeeze().lower()):
            warnings.warn(f'"WS_ms_Avg" is expected to be in m/s!')
        return df

    def run(self):
        version = self.__getVersion()

        if (1 == version):
            warnings.warn("CampbellSciReader (v1)")
            df = self.__readV1()
        elif (2 == version):
            df = self.__readV2()
        else:
            raise Exception("Unreachable instruction!")

        df["sensorFamily"] = "cr1000x"
        df.rename(columns={"TIMESTAMP": "timestamp"}, inplace=True)

        if (df.duplicated().any()):
            _df = df.drop_duplicates(subset=None, keep="first", ignore_index=True, inplace=False)
            msg = f"There is at least one duplicate in {self.inputFile} (from {len(df)} to {len(_df)} rows)"
            warnings.warn(msg)
            df = _df
        if not df.timestamp.is_monotonic_increasing:
            msg = f"Timestamps in {self.inputFile} are not monotonically increasing"
            warnings.warn(msg)

        df.timestamp = df.timestamp.apply(lambda dt: dt.replace(tzinfo=self.tzinfo)) 

        return df

"""
inputFile = "/home/tleduc/prj/coolscapes-tl/dev/data/05-Quai-de-plantes-Nantes/20200518/CS/CR1000XSeries_TwoSec.dat"
df = CampbellSciReader(inputFile).run()
print(df.head(5))

inputFile = "/home/tleduc/prj/nm-ilots-frais/terrain/220711/CR1000XSeries_TwoSec.dat"
df = CampbellSciReader(inputFile).run()
print(df.head(5))

inputFile = "/home/tleduc/prj/prairie-duc/230503/CR1000XSeries_TwoSec.dat"
df = CampbellSciReader(inputFile).run()
print(df.head(5))
"""



'''
Created on 5 oct. 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from rasterio.io import DatasetReader
from rasterio.transform import Affine, AffineTransformer
from shapely import Point
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class RTFromRasterToGeoDataFrameOfPoints(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, raster, oFieldname="value"):
        '''
        Constructor
        '''
        if not isinstance(raster, DatasetReader):
            raise IllegalArgumentTypeException(raster, "DatasetReader")
        self.raster = raster
        self.oFieldname = oFieldname

    @staticmethod
    def __vectorize(raster, fieldname):
        transform, crs = raster.transform, raster.crs
        transformer = AffineTransformer(Affine(*transform))
        pixels = raster.read(1, masked=True)
        nrows, ncols = pixels.shape

        rows = []
        for nr in range(nrows):
            for nc in range(ncols):
                rows.append({
                    "geometry": Point(transformer.xy(nr, nc)),
                    fieldname: pixels[nr, nc]
                })
        return GeoDataFrame(rows, crs=crs)

    def run(self):
        return RTFromRasterToGeoDataFrameOfPoints.__vectorize(self.raster, self.oFieldname)

"""
from numpy import zeros
from t4gpd.demos.GeoDataFrameDemos import GeoDataFrameDemos
from t4gpd.morph.STBBox import STBBox
from t4gpd.raster.RTFromArrayToRaster import RTFromArrayToRaster
from t4gpd.raster.RTToFile import RTToFile
from t4gpd.raster.STRasterize import STRasterize

buildings = GeoDataFrameDemos.regularGridOfPlots(2, 2)
buildings["HAUTEUR"] = 1 + buildings.gid
buildings.to_file("/tmp/buildings.shp")
roi = STBBox(buildings, buffDist=0).run()

dtm = zeros([10, 10])
dtm = RTFromArrayToRaster(dtm, roi).run()
RTToFile(dtm, "/tmp/dtm.tif").run()

dsm = STRasterize(buildings, dtm, attr="HAUTEUR").run()
RTToFile(dsm, "/tmp/dsm.tif").run()

gdf = RTFromRasterToGeoDataFrameOfPoints(dsm).run()
gdf.to_file("/tmp/points.shp")
"""



'''
Created on 11 mai 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import isnan
from pandas.core.frame import DataFrame
from t4gpd.comfort.indices.AbstractThermalComfortIndice import AbstractThermalComfortIndice
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class ASV(AbstractThermalComfortIndice):
    '''
    classdocs
    '''

    def __init__(self, sensorsGdf, AirTC='AirTC_Avg', RH='RH_Avg', WS_ms='WS_ms_Avg',
                 SRUp='SR01Up_3_Avg'):
        '''
        Constructor

        ASV: Actual Sensetion Vote after Nikolopoulou (2004) stated in Coccolo et al. (2016) [-]

        AirTC: air temperature [C]
        RH: relative humidity [%]
        Ws_ms: wind speed recorded at pedestrian level (at height 1.1 m) [m.s-1]
        SRUp: incoming shortwave radiation [W.m-2]
        '''
        if not isinstance(sensorsGdf, DataFrame):
            raise IllegalArgumentTypeException(sensorsGdf, 'DataFrame')

        for fieldname in (AirTC, RH, WS_ms, SRUp):
            if fieldname not in sensorsGdf:
                raise Exception('%s is not a relevant field name!' % fieldname)

        self.AirTC = AirTC
        self.RH = RH
        self.WS_ms = WS_ms
        self.SRUp = SRUp

    def runWithArgs(self, row):
        AirTC = row[self.AirTC]
        RH = row[self.RH]
        WS_ms = row[self.WS_ms]
        SRUp = row[self.SRUp]

        ASV = None
        if not (isnan(AirTC) or isnan(RH) or isnan(WS_ms) or isnan(SRUp)):
            # ASV: Actual Sensetion Vote after (Grosdemouge, 2020, p.105)
            # https://tel.archives-ouvertes.fr/tel-03123710
            ASV = 0.049 * AirTC + 0.001 * SRUp - 0.051 * WS_ms + 0.014 * RH - 2.079
    
            # ASV: Actual Sensetion Vote after Nikolopoulou (2004) stated in Coccolo et al. (2016) [-]
            # ASV = 0.068 * AirTC + 0.0006 * SRUp - 0.107 * WS_ms - 0.002 * RH - 0.69

        return { 'ASV': ASV }



'''
Created on 20 janv. 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from warnings import warn

from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import box, Point
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class LatLonLib(object):
    '''
    classdocs
    '''
    ALGER = GeoDataFrame(
        [{"geometry": Point((3.042048, 36.752887))}], crs="epsg:4326")
    CAP_TOWN = GeoDataFrame(
        [{"geometry": Point((18.42322, -33.92584))}], crs="epsg:4326")
    DAKAR = GeoDataFrame(
        [{"geometry": Point((-17.444061, 14.6937))}], crs="epsg:4326")
    EDINBURGH = GeoDataFrame(
        [{"geometry": Point((-3.188267, 55.953252))}], crs="epsg:4326")
    GREENWICH = GeoDataFrame(
        [{"geometry": Point((-0.001545, 51.477928))}], crs="epsg:4326")
    LAGOS = GeoDataFrame(
        [{"geometry": Point((3.379206, 6.524379))}], crs="epsg:4326")
    # outside the polar circles ; latitude must be between -66.56 and +66.56
    # LONGYEARBYEN = GeoDataFrame([{"geometry": Point((-15.6258907, 78.2253594))}], crs="epsg:4326")
    MONTPELLIER = GeoDataFrame(
        [{"geometry": Point((3.87723, 43.61092))}], crs="epsg:4326")
    NANTES = GeoDataFrame(
        [{"geometry": Point((-1.55, 47.2))}], crs="epsg:4326")
    NIAMEY = GeoDataFrame(
        [{"geometry": Point((2.1098, 13.51366))}], crs="epsg:4326")
    QUITO = GeoDataFrame(
        [{"geometry": Point((-78.52495, -0.22985))}], crs="epsg:4326")
    REYKJAVIK = GeoDataFrame(
        [{"geometry": Point((-21.89541, 64.13548))}], crs="epsg:4326")

    @staticmethod
    def fromGeoDataFrameToLatLon(gdf=NANTES):
        if not isinstance(gdf, GeoDataFrame):
            raise IllegalArgumentTypeException(gdf, "GeoDataFrame")

        if (gdf.crs is None):
            warn("Crs is not set!")
            gdf = LatLonLib.NANTES

        _gdf = gdf if ("epsg:4326" == gdf.crs) else gdf.to_crs("epsg:4326")
        _centroid = box(*_gdf.total_bounds).centroid
        _latitude, _longitude = _centroid.y, _centroid.x
        return _latitude, _longitude



'''
Created on 4 mars 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from io import StringIO
from re import sub

import matplotlib.pyplot as plt
from numpy import float64, linspace, ndarray, vectorize
from pandas import DataFrame, read_csv
from scipy.stats import norm
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class AbstractEnergyDemand(object):
    '''
    classdocs
    '''

    def __init__(self, D=None, S=None, E=None):
        '''
        Constructor
        '''
        if D is None:
            _D = self._getDevicesPerAgeGroup()
            self.D = _D.drop(columns=['device']).to_numpy(dtype=float64)
            # self.D = _D.drop(columns=['device']).to_numpy(dtype=str)
        elif isinstance(D, ndarray):
            self.D = D
        elif isinstance(D, DataFrame):
            self.D = D.to_numpy()
        else:
            raise IllegalArgumentTypeException(D, 'numpy.ndarray or pandas.DataFrame')

        if S is None:
            _S = self._getNormalScenario()
            self.S = _S.drop(columns=['Age group']).to_numpy(dtype=str)
        elif isinstance(S, ndarray):
            self.S = S
        elif isinstance(S, DataFrame):
            self.S = S.to_numpy()
        else:
            raise IllegalArgumentTypeException(S, 'numpy.ndarray or pandas.DataFrame')

        if E is None:
            _E = self._getDeviceEnergyConsumption()
            self.E = _E.drop(columns=['device', 'comment']).to_numpy(dtype=str)
        elif isinstance(E, ndarray):
            self.E = E
        elif isinstance(E, DataFrame):
            self.E = E.to_numpy()
        else:
            raise IllegalArgumentTypeException(E, 'numpy.ndarray or pandas.DataFrame')

        # CHECK DIMENSIONS
        n, m = self.D.shape

        if not (((1 == self.S.ndim) and (m == self.S.shape[0])) or
                ((2 == self.S.ndim) and (m == self.S.shape[0]) and  (1 == self.S.shape[1]))):
            raise IllegalArgumentTypeException(self.S, f'1D numpy.ndarray or 2D numpy.ndarray (m={m} rows x 1 col.)')

        if not (((1 == self.E.ndim) and (n == self.E.shape[0])) or
                ((2 == self.E.ndim) and (n == self.E.shape[0]) and  (1 == self.E.shape[1]))):
            raise IllegalArgumentTypeException(self.S, f'1D numpy.ndarray or 2D numpy.ndarray (n={3} rows x 1 col.)')

        self.S = self.S.reshape(-1)
        self.E = self.E.reshape(-1)

    @staticmethod
    def _getDevicesPerAgeGroup():
        _sio = StringIO("""device;Men;Ind_11_17;Ind_18_24;Ind_25_39;Ind_40_54;Ind_55_64;Ind_65_79;Ind_80p
gateway;95.0;0.0;0.0;0.0;0.0;0.0;0.0;0.0
laptop;0.0;91.0;82.0;73.0;83.0;76.0;58.0;0.0
tablet;0.0;37.0;36.0;49.0;45.0;47.0;29.0;0.0
smartphone;0.0;86.0;98.0;95.0;80.0;62.0;44.0;0.0
smart object;0.0;31.0;25.0;21.0;17.0;9.0;5.0;0.0
smart speaker;0.0;15.0;14.0;1.0;9.0;7.0;5.0;0.0
tv;0;0;79.2;93.1;95.3;97.9;97.9;0.0
""")
        return read_csv(_sio, sep=';')

    @staticmethod
    def _getDeviceEnergyConsumption():
        _sio = StringIO("""device;Wh;comment
gateway;26.0;150 kWh/an < box < 300 kWh/an
laptop;75.0;50 Wh < laptop < 100 Wh
tablet;4.0;2 Wh < tablet < 6 Wh
smartphone;10*delta(h-19);Once a day
smart object;1.0;
smart speaker;5.0;
tv;100.0;40 Wh (Oled) < TV < 250 Wh (Plasma)
""")
        return read_csv(_sio, sep=';')

    @staticmethod
    def _getNormalScenario():
        '''
        _sio = StringIO("""Age group;IT equipment usage rate
Men;100
Ind_11_17;rect(0,5,w) * (rect(12,14,h) * 20 + rect(18,20,h) * 70 + rect(20,23,h) * 50) + rect(5,7,w) * (rect(8,13,h) * 90 + rect(18,20,h) * 70 + rect(20,22,h) * 90)
Ind_18_24;rect(0,5,w) * (rect(19,21,h) * 70 + rect(21,24,h) * 60) + rect(5,7,w) * (rect(10,15,h) * 50 + rect(21,24,h) * 70)
Ind_25_39;rect(0,5,w) * (rect(9,12,h) * 20 + rect(12,14,h) * 30 + rect(19,21,h) * 60 + rect(21,24,h) * 70) + rect(5,7,w) * (rect(10,15,h) * 90 + rect(18,24,h) * 90)
Ind_40_54;rect(0,5,w) * (rect(9,12,h) * 30 + rect(12,14,h) * 35 + rect(19,21,h) * 80 + rect(21,23,h) * 70) + rect(5,7,w) * (rect(8,10,h) * 50 + rect(12,14,h) * 60 + rect(18,22,h) * 70)
Ind_55_64;rect(0,5,w) * (rect(9,12,h) * 40 + rect(12,14,h) * 35 + rect(19,21,h) * 80) + rect(5,7,w) * (rect(7,9,h) * 90 + rect(12,14,h) * 60 + rect(18,22,h) * 70)
Ind_65_79;rect(7,9,h) * 20 + rect(9,12,h) * 40 + rect(12,15,h) * 90 + rect(15,17,h) * 50 + rect(17,22,h) * 70
Ind_80p;rect(9,13,h) * 60 + rect(14,22,h) * 40
""")
        '''
        _sio = StringIO("""Age group;IT equipment usage rate
Men;100
Ind_11_17;rect(0,5,w) * (rect(12,14,h) * snd(20) + rect(18,20,h) * snd(70) + rect(20,23,h) * snd(50)) + rect(5,7,w) * (rect(8,13,h) * snd(90) + rect(18,20,h) * snd(70) + rect(20,22,h) * snd(90))
Ind_18_24;rect(0,5,w) * (rect(19,21,h) * snd(70) + rect(21,24,h) * snd(60)) + rect(5,7,w) * (rect(10,15,h) * snd(50) + rect(21,24,h) * snd(70))
Ind_25_39;rect(0,5,w) * (rect(9,12,h) * snd(20) + rect(12,14,h) * snd(30) + rect(19,21,h) * snd(60) + rect(21,24,h) * snd(70)) + rect(5,7,w) * (rect(10,15,h) * snd(90) + rect(18,24,h) * snd(90))
Ind_40_54;rect(0,5,w) * (rect(9,12,h) * snd(30) + rect(12,14,h) * snd(35) + rect(19,21,h) * snd(80) + rect(21,23,h) * snd(70)) + rect(5,7,w) * (rect(8,10,h) * snd(50) + rect(12,14,h) * snd(60) + rect(18,22,h) * snd(70))
Ind_55_64;rect(0,5,w) * (rect(9,12,h) * snd(40) + rect(12,14,h) * snd(35) + rect(19,21,h) * snd(80)) + rect(5,7,w) * (rect(7,9,h) * snd(90) + rect(12,14,h) * snd(60) + rect(18,22,h) * snd(70))
Ind_65_79;rect(7,9,h) * snd(20) + rect(9,12,h) * snd(40) + rect(12,15,h) * snd(90) + rect(15,17,h) * snd(50) + rect(17,22,h) * snd(70)
Ind_80p;rect(9,13,h) * snd(60) + rect(14,22,h) * snd(40)
""")
        return read_csv(_sio, sep=';')

    @staticmethod
    def _unitImpulse(x):
        return 1 if (0 == x) else 0

    @staticmethod
    def _indicatorFunction(x0, x1, x):
        return 1 if (x0 <= x < x1) else 0

    @staticmethod
    def _randomVariates(mu, std=None):
        if std is None:
            std = 0.1 * mu
        return norm(mu, std).rvs()

    def __eval(self, ch):
        h, w = self.h, self.w
        rect = lambda t0, t1, t: self._indicatorFunction(t0, t1, t)
        snd = lambda x: x
        return eval(ch) if isinstance(ch, str) else ch 

    def plotScenario(self, ofile=None):
        _eval_vector = vectorize(self.__eval)

        AGE_GROUPS = ['Men.', '11_17', '18-24', '25-39', '40-54', '55-64', '65-79', '80+']
        COLORS = ['darkgrey', 'dimgrey', 'black']
        STYLES = ['solid', 'dashed', 'dashdot', 'dotted']
        MARKERS = ['+', 'h', 'D', '', 'o', '>']
        WIDTH = [2, 3]
        H = range(24)

        for day_label, self.w in [('weekday', 0), ('weekend', 5)]:
            rows = []
            for self.h in H:
                rows.append(_eval_vector(self.S))
            df = DataFrame(rows, columns=AGE_GROUPS)
            df['H'] = H

            fig, ax = plt.subplots(figsize=(1 * 8.26, 1 * 8.26))
            # ax.set_title(day_label, size=24)

            for i, ageGroup in enumerate(AGE_GROUPS[1:]):
                ax.scatter(H, df[ageGroup], color=COLORS[i % 3], marker=MARKERS[i % len(MARKERS)])
                ax.plot(H, df[ageGroup], color=COLORS[i % 3],
                        linestyle=STYLES[i % 4], linewidth=WIDTH[i % 2])
                ax.plot([], [], color=COLORS[i % 3], marker=MARKERS[i % len(MARKERS)],
                        linestyle=STYLES[i % 4], linewidth=WIDTH[i % 2], label=ageGroup)

            ax.tick_params(axis='both', which='major', labelsize=16)
            ax.tick_params(axis='both', which='minor', labelsize=10)
            ax.set_ylim(0, 100)
            ax.set_xlabel('Hour', size=20)
            ax.set_ylabel('IT equipment usage ratios (in %)', size=20)
            ax.legend(prop={'size': 20})

            if ofile is None:
                plt.show()
            else:
                plt.savefig(f'{day_label}.pdf', bbox_inches='tight')
            plt.close(fig)

    @staticmethod
    def to_latex():
        _COLUMNS = {'Men': 'household', 'Ind_11_17': '11-17', 'Ind_18_24': '18-24', 'Ind_25_39': '25-39',
                   'Ind_40_54': '40-54', 'Ind_55_64': '55-64', 'Ind_65_79': '65-79', 'Ind_80p': '80+'}

        _D = AbstractEnergyDemand._getDevicesPerAgeGroup()
        _D = _D.rename(columns=_COLUMNS)
        _D = _D.astype(str)

        _E = AbstractEnergyDemand._getDeviceEnergyConsumption()
        _E = _E.drop(columns=['comment']).set_index('device')\
            .rename(columns={'Wh': 'Energy consumption'}).transpose()
        _E['smartphone'] = _E['smartphone'].apply(
            lambda s: s.replace('*delta', '$\\times\\delta$'))

        _S = AbstractEnergyDemand._getNormalScenario()
        _S = _S.set_index('Age group').transpose().rename(columns=_COLUMNS).transpose().reset_index()
        _S['IT equipment usage rate'] = _S['IT equipment usage rate'].apply(
            lambda s: s.replace('rect', '$\\Pi$'))
        _S['IT equipment usage rate'] = _S['IT equipment usage rate'].apply(
            lambda s: s.replace('*', '$\\times$'))
        '''
        _S['IT equipment usage rate'] = _S['IT equipment usage rate'].apply(
            lambda s: s.replace('snd', '$\\Phi$'))
        '''
        _S['IT equipment usage rate'] = _S['IT equipment usage rate'].apply(
            lambda s: sub('snd\(([0-9]*)\)', '$Phi_{\\1}$', s).replace('Phi', '\\Phi'))

        print(f'''
{_D.style.hide().to_latex(hrules=True)}

{_E.style.to_latex(hrules=True)}

{_S.style.hide().to_latex(hrules=True)}
''')



'''
Created on 22 aug. 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import rasterio
from geopandas import GeoDataFrame
from numpy import asarray, ndarray
from rasterio.io import DatasetReader
from rasterio.shutil import copy
from rasterio.transform import Affine
from tempfile import TemporaryDirectory
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.raster.AbstractRasterGeoProcess import AbstractRasterGeoProcess


class RTFromArrayToRaster(AbstractRasterGeoProcess):
    '''
    classdocs
    '''

    def __init__(self, array, rasterOrRoi, debug=False):
        '''
        Constructor
        '''
        if not isinstance(array, (list, ndarray, tuple)):
            raise IllegalArgumentTypeException(
                array, "list, numpy ndarray, or tuple")
        self.array = asarray(array).astype("float32")
        if (1 == self.array.ndim):
            self.array = self.array.reshape(1, -1)
        if (2 != self.array.ndim):
            raise IllegalArgumentTypeException(
                array, "*1D* or *2D* list, numpy ndarray, or tuple")

        if not isinstance(rasterOrRoi, (DatasetReader, GeoDataFrame)):
            raise IllegalArgumentTypeException(
                rasterOrRoi, "DatasetReader or GeoDataFrame")
        self.rasterOrRoi = rasterOrRoi
        self.debug = debug

    @staticmethod
    def __extract_metadata_from_DatasetReader(array, raster):
        if (array.shape[0] != raster.height):
            raise Exception(
                f"array.shape[0] ({array.shape[0]}) <> raster.height ({raster.height})!")
        if (array.shape[1] != raster.width):
            raise Exception(
                f"array.shape[1] ({array.shape[1]}) <> raster.width ({raster.width})!")

        out_meta = raster.meta.copy()
        out_meta.update({
            "driver": "GTiff",
            "height": raster.height,
            "width": raster.width,
            "transform": raster.transform,
            "crs": raster.crs,
            "count": 1,
            "dtype": array.dtype,
            "nodata": -9999
        })
        return out_meta

    @staticmethod
    def __extract_metadata_from_GeoDataFrame(array, roi):
        minx, miny, maxx, maxy = roi.total_bounds
        nrows, ncols = array.shape

        xres = (maxx-minx) / ncols
        yres = (maxy-miny) / nrows

        transform = Affine.translation(minx, maxy) * Affine.scale(xres, -yres)

        out_meta = {
            "driver": "GTiff",
            "height": nrows,
            "width": ncols,
            "transform": transform,
            "crs": roi.crs,
            "count": 1,
            "dtype": array.dtype,
            "nodata": -9999,
            # "blockxsize": 1,
            # "blockysize": 1,
            # "compress": "lzw",
        }
        return out_meta

    def run(self):
        if isinstance(self.rasterOrRoi, DatasetReader):
            out_meta = RTFromArrayToRaster.__extract_metadata_from_DatasetReader(
                self.array, self.rasterOrRoi)

        elif isinstance(self.rasterOrRoi, GeoDataFrame):
            out_meta = RTFromArrayToRaster.__extract_metadata_from_GeoDataFrame(
                self.array, self.rasterOrRoi)

        result = self._write_and_load(self.array, out_meta, self.debug)
        return result


"""
import contextily as ctx
import matplotlib.pyplot as plt
from t4gpd.raster.RTToFile import RTToFile
from t4gpd.demos.GeoDataFrameDemos import GeoDataFrameDemos
from t4gpd.morph.STBBox import STBBox
from numpy import arange

buildings = GeoDataFrameDemos.districtGraslinInNantesBuildings()
roi = STBBox(buildings, buffDist=0).run()

buildings.to_file("/tmp/buildings.shp")
roi.to_file("/tmp/roi.shp")

array = arange(20).reshape(4, -1).astype(float)
array = arange(20).reshape(2, -1).astype(float)
raster1 = RTFromArrayToRaster(array, roi).run()
RTToFile(raster1, "/tmp/raster1.tif").run()

fig, ax = plt.subplots(figsize=(1 * 8.26, 1 * 8.26))
buildings.boundary.plot(ax=ax, color="red", hatch=".")
ctx.add_basemap(ax, crs=buildings.crs, source="/tmp/raster1.tif")
ax.axis("off")
plt.tight_layout()
plt.show()
"""

"""
from geopandas import read_file
from t4gpd.raster.RTClip import RTClip
from t4gpd.raster.RTLoad import RTLoad
from t4gpd.raster.RTToFile import RTToFile
from t4gpd.raster.STRasterize import STRasterize

srcdir = "/home/tleduc/prj/uenv_aau_msc_course/tps"
dtmdir = "/home/tleduc/data/mnt_asc/mnt_nantes_2005_1m";
raster = RTLoad(f"{dtmdir}/MNT_L93_0354_6690.asc").run()

#~ CROP RASTER FILE
roi = read_file(f"{srcdir}/data/roi.shp")
dtm = RTClip(raster, roi).run()
RTToFile(dtm, f"{srcdir}/data/dtm.tif").run()

#~ BUILDINGS LAYER RASTERIZING
buildings = read_file(f"{srcdir}/data/buildings.shp")
_dsm1 = STRasterize(buildings, dtm, attr="HAUTEUR").run()
RTToFile(_dsm1, f"{srcdir}/data/_dsm1.tif").run()

#~ STATUE LAYER RASTERIZING
statue = read_file(f"{srcdir}/data/statue.shp")
_dsm2 = STRasterize(statue, dtm, attr="HAUTEUR").run()
RTToFile(_dsm2, f"{srcdir}/data/_dsm2.tif").run()

#~ DSM1 + DSM2 => DSM12
dsm12 = _dsm1.read(1) + _dsm2.read(1)
dsm12 = RTFromArrayToRaster(dsm12, roi).run()
RTToFile(dsm12, f"{srcdir}/data/dsm12.tif").run()
"""



'''
Created on 3 oct 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import exp


class Angstrom(object):
    '''
    classdocs

    http://dx.doi.org/10.1016/j.buildenv.2014.05.019
    Stephane Boltzmann's constant [W . m^-2 . K^-4]
    '''

    BOLTZMANN_CONSTANT = 5.67e-8

    @staticmethod
    def __vapor_pressure(Tair, RH):
        '''
        IN:
            Tair: Dry bulb temperature [C]
            RH: Relative humidity 0-100 [%]
        OUT:
            Water vapor pressure [hPa]

        http://dx.doi.org/10.1016/j.buildenv.2014.05.019

        section 2.2: Long-wave radiation from the atmosphere
        The vapor pressure Vp [hPa] can be computed using the Arden Buck equation
        '''
        return RH * 6.1121 * exp((18.678 - Tair / 234.4) * (Tair / (257.14 + Tair)))

    @staticmethod
    def lowg_wave_irradiance(Tair, RH, N):
        '''
        IN:
            Tair: Dry bulb temperature [C]
            RH: Relative humidity 0-100 [%]
            N: Degree of cloudiness [Octas]
        OUT:
            Downward atmospheric long-wave irradiance [W . m^-2]

        http://dx.doi.org/10.1016/j.buildenv.2014.05.019

        section 2.2: Long-wave radiation from the atmosphere
        Atmospheric long-wave radiation can be estimated as a function of ambient air
        temperature, vapor pressure, and cloud cover using the Angstrom formula

        Sky conditions are estimated in terms of how many eighths of the sky are
        covered in cloud, ranging from 0 Octas (completely clear sky) through to
        8 Octas (completely overcast)
        '''
        Vp = Angstrom.__vapor_pressure(Tair, RH)
        return Angstrom.BOLTZMANN_CONSTANT * \
            (Tair + 273.15) ** 4 * \
            (0.82 - 0.25 * 10 ** (-0.0945 * Vp)) * \
            (1 + 0.21 * (N/8) ** 2.5)



'''
Created on 12 jav. 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from numpy import cos, hstack, linspace, pi, sin 
from pandas import DataFrame
from shapely.geometry import Polygon
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException

import matplotlib as mpl, matplotlib.pyplot as plt


class RoseMappingTool(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, df, label='label', color='color', solid=False, power=10, radius=1, npts=75):
        '''
        Constructor
        '''
        if not isinstance(df, DataFrame):
            raise IllegalArgumentTypeException(df, 'DataFrame')
        self.df = df

        if not label in df:
            raise Exception('%s is not a relevant field name!' % label)
        self.label = label

        if not (color is None or (color in df)):
            raise Exception('%s is not a relevant field name!' % color)
        self.color = color

        if not (isinstance(solid, bool) or (solid in df)):
            raise Exception('%s is not a relevant field name!' % solid)
        self.solid = solid

        self.npetals = len(df)
        self.power = power
        self.radius = radius
        self.npts = npts

    def _petal_old(self, npetal, npts=100):
        # npetal: 0 -> self.npetals - 1
        deltaTheta = pi / (2 * self.npetals)

        if (0 == npetal):
            n0, n1 = 0, 2 * len(self.df) - 1
        else:
            n0, n1 = 2 * npetal - 1, 2 * npetal

        theta = hstack([
            linspace(n0 * deltaTheta, (n0 + 1) * deltaTheta, npts),
            linspace(n1 * deltaTheta, (n1 + 1) * deltaTheta, npts),
            ])
        x = self.radius * cos(self.npetals * theta) * cos(theta)
        y = self.radius * cos(self.npetals * theta) * sin(theta)

        return x, y

    def _petal(self, npetal, npts=100):
        # npetal: 0 -> self.npetals - 1
        deltaTheta = pi / self.npetals

        if (0 == npetal):
            n0, n1 = 0, 2 * len(self.df) - 1
        else:
            n0, n1 = 2 * npetal - 1, 2 * npetal

        theta = hstack([
            linspace(n0 * deltaTheta, (n0 + 1) * deltaTheta, npts),
            linspace(n1 * deltaTheta, (n1 + 1) * deltaTheta, npts),
            ])
        _tmp = self.radius * (1 - sin((self.npetals / 2) * theta) ** self.power)
        x = _tmp * cos(theta - pi / 2)
        y = _tmp * sin(theta - pi / 2)

        return x, y

    @staticmethod
    def plot(gdf, fontsize=13, hatch='x', hatchwidth=0.1, oFile=None):
        foo = lambda s: '\n'.join(s.split(' '))

        fig, ax = plt.subplots(figsize=(8.26, 8.26))
        for solid, color in zip(gdf.solid, gdf.color):
            _gdf = gdf[gdf.color == color]
            if solid:
                _gdf.plot(ax=ax, color=color, alpha=0.5)
                _gdf.apply(lambda x: ax.annotate(
                    text=foo(x.label), xy=x.geometry.centroid.coords[0],
                    color='black', size=fontsize, weight='bold',
                    ha='center', va='center'), axis=1);
            else:
                mpl.rcParams['hatch.linewidth'] = hatchwidth
                _gdf.boundary.plot(ax=ax, color=color, hatch=hatch)
                _gdf.apply(lambda x: ax.annotate(
                    text=foo(x.label), xy=x.geometry.centroid.coords[0],
                    color=color, size=fontsize, weight='bold',
                    ha='center', va='center'), axis=1);
        ax.axis('off')
        if oFile is None:
            plt.show()
        else:
            plt.savefig(oFile, bbox_inches='tight')            
        plt.close(fig)
        
    def run(self):
        # ~ https://fr.wikipedia.org/wiki/Rosace_(math%C3%A9matiques)
        rows = []
        for npetal, row in self.df.iterrows():
            color, label = row[self.color], row[self.label]
            solid = self.solid if isinstance(self.solid, bool) else row[self.solid]
            x, y = self._petal(npetal, self.npts)
            rows.append({
                'npetal': npetal,
                'geometry': Polygon([(_x, _y) for _x, _y in zip(x, y)]),
                'label': label,
                'color': color,
                'solid': solid
                })
        return GeoDataFrame(rows)

'''
from io import StringIO
from pandas import read_csv

_sio = StringIO("""label;color;solid
Géographie, aménagement et développement;red;True
Études urbaines;grey;False
Énergies renouvelables, développement durable et environnement;green;False
Sciences de l'environnement;orange;True
Informatique;magenta;False
""")
df = read_csv(_sio, sep=';')

gdf = RoseMappingTool(df, label='label', color='color', solid='solid', power=10).run()
print(gdf)
RoseMappingTool.plot(gdf, fontsize=12, hatch='x',
                     hatchwidth=0.1, oFile=None)
'''



'''
Created on 22 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from t4gpd.commons.Distances import EuclideanDistance


class AbstractMethod(object):
    '''
    classdocs
    '''

    def __init__(self, params):
        '''
        Constructor
        '''
        raise NotImplementedError('The class AbstractMethod can not be instanciated!')

    def attrName(self):
        raise NotImplementedError('The method must be implemented!')

    def distance(self, signal, signalRef):
        return EuclideanDistance.compute(signal, signalRef)
        # return np.sum(abs(signal - signalRef)) / np.sum(abs(signalRef))
        # return np.sum((signal - signalRef)**2) / np.sum(signalRef**2)

    def nearestPattern(self, shapeSignature, patternSignatures):
        minDist, closestPatternID = float('inf'), None

        for gid, patternSignature in list(patternSignatures.items()):
            _dist = self.distance(shapeSignature, patternSignature)
            if _dist < minDist:
                minDist = _dist
                closestPatternID = gid
        return { self.attrName():closestPatternID }

    def signature(self, raylens, isAPatternLayer):
        raise NotImplementedError('The method must be implemented!')



'''
Created on 21 sep. 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import matplotlib.pyplot as plt
import matplotlib.ticker as tck
from datetime import datetime, timedelta, timezone
from locale import LC_ALL, setlocale
from numpy import asarray, sqrt
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.LatLonLib import LatLonLib
from t4gpd.commons.sun.SunLib import SunLib
from t4gpd.energy.DirectSolarIrradianceLib import DirectSolarIrradianceLib


class FelixMarboutin(GeoProcess):
    '''
    classdocs

    Irradiance is understood as instantaneous density of solar radiation incident on a given
    surface, typically expressed in W/m2.

    Irradiation is the sum of irradiance over a time period (e.g. 1 hour, day, month, etc.)
    expressed in J/m2. However, in daily routine Wh/m2 are more commonly used.
    '''
    R = sqrt(2)/2
    NORMALS = {
        "Roof": {"normalvector": (0, 0, 1), "linestyle": "dashdot", "linewidth": 2.25, "color": "black"},
        "Ground": {"normalvector": (0, 0, -1), "linestyle": "solid", "linewidth": 2.25, "color": "black"},
        "N": {"normalvector": (0, 1, 0), "linestyle": "solid", "linewidth": 2.25, "color": "blue"},
        "S": {"normalvector": (0, -1, 0), "linestyle": "solid", "linewidth": 2.25, "color": "red"},
        "E": {"normalvector": (1, 0, 0), "linestyle": "dotted", "linewidth": 2.75, "color": "darkorange"},
        "W": {"normalvector": (-1, 0, 0), "linestyle": "dashed", "linewidth": 2.75, "color": "brown"},
        "NE": {"normalvector": (R, R, 0), "linestyle": "dashed", "linewidth": 2.25, "color": "blue"},
        "NW": {"normalvector": (-R, R, 0), "linestyle": "dashed", "linewidth": 2.25, "color": "red"},
        "SW": {"normalvector": (-R, -R, 0), "linestyle": "dotted", "linewidth": 2.25, "color": "darkorange"},
        "SE": {"normalvector": (R, -R, 0), "linestyle": "dotted", "linewidth": 2.25, "color": "brown"},
    }
    YEAR = datetime.today().year
    DMJ = [
        datetime(YEAR, 12, 21, tzinfo=timezone.utc),
        datetime(YEAR, 3, 21, tzinfo=timezone.utc),
        datetime(YEAR, 6, 21, tzinfo=timezone.utc),
    ]
    MINUTES = range(0, 60, 15)

    def __init__(self, position=LatLonLib.NANTES,
                 normals=["Roof", "N", "S", "E", "W"],
                 stepInDays=10,
                 model="pysolar"):
        '''
        Constructor
        '''
        if ("epsg:4326" != position.crs):
            raise IllegalArgumentTypeException(
                position, "Must be a WGS84 GeoDataFrame")
        self.position = position
        self.lat, _ = LatLonLib.fromGeoDataFrameToLatLon(position)
        self.sunModel = SunLib(position, model=model)

        if not (set(normals) <= set(FelixMarboutin.NORMALS.keys())):
            raise IllegalArgumentTypeException(
                normals, f"Must be contained in {FelixMarboutin.NORMALS.keys()}")
        self.normals = normals
        self.stepInDays = stepInDays

    def __timeSinceMidnight(dt):
        midnight = datetime(dt.year, dt.month, dt.day, tzinfo=timezone.utc)
        return (dt - midnight).seconds / 3600

    def __dailyIrradiances(self, dt0, sunrise, sunset, normal):
        X, Y = [], []
        for hour in range(int(sunrise), int(sunset)+1):
            for minutes in FelixMarboutin.MINUTES:
                dt = dt0 + timedelta(hours=hour, minutes=minutes)
                di = DirectSolarIrradianceLib.noMaskDI(
                    normal, dt, gdf=self.position)
                X.append(hour + minutes/60)
                Y.append(di)
        return X, Y

    def daily(self, dts=DMJ, ofile=None):
        sunrise = min([FelixMarboutin.__timeSinceMidnight(
            self.sunModel.getSunrise(dt)) for dt in dts])
        sunset = max([FelixMarboutin.__timeSinceMidnight(
            self.sunModel.getSunset(dt)) for dt in dts])

        ncols = len(dts)
        fig, axes = plt.subplots(ncols=ncols, figsize=(
            ncols * 8.26, 8.26), squeeze=False)
        fig.suptitle(
            f"Direct Solar Irradiance, lat. {self.lat:+.1f}°N", size=20)

        for nfig, dt0 in enumerate(dts):
            ax = axes[0, nfig]

            for normal in self.normals:
                _normal = FelixMarboutin.NORMALS[normal]
                X, Y = self.__dailyIrradiances(
                    dt0, sunrise, sunset, _normal["normalvector"])
                totalIrr = sum(Y) / (len(FelixMarboutin.MINUTES) * 1e3)
                ax.plot(X, Y, label=f"{normal}: {totalIrr:.1f} kWh/m$^2$",
                        linestyle=_normal["linestyle"],
                        linewidth=_normal["linewidth"],
                        color=_normal["color"])

            setlocale(LC_ALL, "en_US.utf8")
            ax.set_title(f"{dt0.strftime('%b %d')}", fontsize=18)
            ax.set_xlabel("Hour (UTC)", fontsize=14)
            ax.legend(loc="upper left", framealpha=0.5, ncol=2, fontsize=14)
            ax.set_ylim([0, 1000])
            ax.xaxis.set_major_locator(tck.MultipleLocator(base=2))
            ax.tick_params(axis="x", labelsize=14)
            ax.tick_params(axis="y", labelsize=14)
            if (0 == nfig):
                ax.set_ylabel(
                    "Instantaneous irradiance [W.m$^{-2}$]", fontsize=14)

        if ofile is None:
            fig.tight_layout()
            plt.show()
        else:
            plt.savefig(ofile, bbox_inches="tight")
        plt.close(fig)

    def __annualIrradiances(self, sunrise, sunset, normal):
        X, Y = range(0, 365, self.stepInDays), []

        dt0 = datetime(FelixMarboutin.YEAR, 1, 1, tzinfo=timezone.utc)
        for yday in X:
            irr = 0
            for hour in range(int(sunrise), int(sunset)+1):
                dt = dt0 + timedelta(days=yday, hours=hour)
                irr += DirectSolarIrradianceLib.noMaskDI(
                    normal, dt, gdf=self.position)
            Y.append(irr)
        return X, Y

    def annual(self, ofile=None):
        w_solstice = datetime(FelixMarboutin.YEAR, 12, 21)
        s_solstice = datetime(FelixMarboutin.YEAR, 6, 21)

        sunrise = min([FelixMarboutin.__timeSinceMidnight(
            self.sunModel.getSunrise(dt)) for dt in [w_solstice, s_solstice]])
        sunset = max([FelixMarboutin.__timeSinceMidnight(
            self.sunModel.getSunset(dt)) for dt in [w_solstice, s_solstice]])

        fig, ax = plt.subplots(figsize=(8.26, 0.55*8.26))

        for normal in self.normals:
            _normal = FelixMarboutin.NORMALS[normal]
            X, Y = self.__annualIrradiances(
                sunrise, sunset, _normal["normalvector"])
            Y = 1e-3 * asarray(Y)

            totalIrr = self.stepInDays * sum(Y) * 1e-3
            ax.plot(X, Y, label=f"{normal}: {totalIrr:.2f} MWh/m$^2$",
                    linestyle=_normal["linestyle"],
                    linewidth=_normal["linewidth"],
                    color=_normal["color"])
            ax.legend(loc="upper left", framealpha=0.5, ncol=2, fontsize=14)

        fig.suptitle(
            f"Direct Solar Irradiation, lat. {self.lat:+.1f}°N", size=20)
        ax.set_xlabel("Day of year", fontsize=14)
        ax.set_ylabel("Daily irradiation [kWh/m$^2$]", fontsize=14)
        ax.tick_params(axis="x", labelsize=14)
        ax.tick_params(axis="y", labelsize=14)

        for month, label in [(3, "Spring equinox"), (6, "Summer solstice"), (9, "Autumn equinox")]:
            x0 = datetime(FelixMarboutin.YEAR, month, 21, 12,
                          tzinfo=timezone.utc).timetuple().tm_yday
            ax.axvline(x=x0, linestyle="dotted", color="blue", linewidth=1)
            ax.text(x0, 2.5, label, rotation=90,
                    color="blue", va="center", fontsize=14)

        if ofile is None:
            fig.tight_layout()
            plt.show()
        else:
            plt.savefig(ofile, bbox_inches="tight")
        plt.close(fig)


"""
fm = FelixMarboutin()
fm.daily()
fm.annual()

FelixMarboutin(
    normals=FelixMarboutin.NORMALS.keys()
).daily(dts=[datetime(2023, 6, 21)])
"""



'''
Created on 12 mai 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import isnan
from pandas.core.frame import DataFrame
from t4gpd.comfort.indices.AbstractThermalComfortIndice import AbstractThermalComfortIndice
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class THI(AbstractThermalComfortIndice):
    '''
    classdocs
    '''

    def __init__(self, sensorsGdf, AirTC='AirTC_Avg', RH='RH_Avg'):
        '''
        Constructor

        AirTC: air temperature [C]
        RH: relative humidity [%]

        THI: Temperature-Humidity Index (THI) stated in Grosdemouge (2020) 
        after (Emmanuel et al., 2016) in humid and hot environment
        *** 21-24: 100% des personnes en situation de confort
        *** 24-26: 50% des personnes en situation de confort
        *** > 26: 100% des personnes en situation d'inconfort lie a la chaleur
        '''
        if not isinstance(sensorsGdf, DataFrame):
            raise IllegalArgumentTypeException(sensorsGdf, 'DataFrame')

        for fieldname in (AirTC, RH):
            if fieldname not in sensorsGdf:
                raise Exception('%s is not a relevant field name!' % fieldname)

        self.AirTC = AirTC
        self.RH = RH

    def runWithArgs(self, row):
        AirTC = row[self.AirTC]
        RH = row[self.RH]

        THI = None
        if not (isnan(AirTC) or isnan(RH)):
            # THI: Temperature-Humidity Index
            THI = 0.8 * AirTC + (AirTC * RH) / 500

        return { 'THI': THI }



'''
Created on 17 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import arctan2, cos, linspace, pi, round, sin
from numpy.random import uniform
from shapely.geometry import LineString, Point, Polygon
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class SequenceRadii(object):
    '''
    classdocs
    '''

    def __init__(self, nbranchs=8, width=12.0, length=100.0, varLength=None):
        '''
        Constructor
        '''
        self.nbranchs = nbranchs
        self.width = 0.5 * float(width)
        self.length = length - self.width
        self.incAngle = arctan2(self.width, self.length)
        self.angles = linspace(0, 2 * pi, num=nbranchs, endpoint=False)
        self.mainDirs = list(zip(cos(self.angles), sin(self.angles)))

        if varLength is None:
            self.lowerBoundary = 1.0
        elif (0 <= varLength <= 1.0):
            self.lowerBoundary = 1.0 - varLength
        else:
            raise IllegalArgumentTypeException(varLength, "Float value between 0.0 and 1.0 or None")

    def getBranch(self, i, centre=[0, 0]):
        length = uniform(low=self.lowerBoundary, high=1) * self.length
        ux, uy = self.mainDirs[i]
        p = (centre[0] + length * ux, centre[1] + length * uy)
        return LineString([centre, p])

    def getSector(self, sector, centre=[0, 0]):
        length = uniform(low=self.lowerBoundary, high=1) * self.length
        i, j = sector

        if ((0 == i) and (self.nbranchs == j)):
            result = Point(centre).buffer(length)
        else:
            cx, cy = centre
            angle0, angle1 = self.angles[i], self.angles[j]
            npts = round((angle1 - angle0) / self.incAngle).astype(int)

            result = [centre]
            for angle in linspace(angle0, angle1, num=npts, endpoint=True):
                p = (cx + length * cos(angle), cy + length * sin(angle))
                result.append(p)
            result.append(centre)
            result = Polygon(result)

        return result

    def getWidth(self):
        return self.width



'''
Created on 28 sept. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import LineString, MultiPolygon, Polygon
from shapely.ops import unary_union
from shapely.prepared import prep
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.morph.STPointsDensifier import STPointsDensifier
from t4gpd.morph.STVoronoiPartition import STVoronoiPartition


class STSkeletonize(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, inputGdf, samplingDist=10.0):
        '''
        Constructor
        '''
        if not isinstance(inputGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(inputGdf, 'GeoDataFrame')
        self.inputGdf = inputGdf

        self.samplingDist = samplingDist

    def __skeletonizePolygon(self, shapeGeom):
        # Use a buffer to avoid slivers
        shapeGeom = shapeGeom.buffer(0.001)
        prepShapeGeom = prep(shapeGeom)

        # Sample the building contours
        _currGdf = GeoDataFrame([{ 'geometry': shapeGeom }])
        _nodesGdf = STPointsDensifier(_currGdf, self.samplingDist).run()

        if (5 < len(_nodesGdf)):
            # Voronoi Diagram
            _voronoiGdf = STVoronoiPartition(_nodesGdf).run()
            lls = []
            for geom in _voronoiGdf.geometry:
                lls.extend(filter(prepShapeGeom.contains, GeomLib.toListOfBipointsAsLineStrings(geom)))
            return unary_union(lls)

        return LineString()  # Empty geometry

    def run(self):
        rows = []
        for _, row in self.inputGdf.iterrows():
            geom = row.geometry
            if isinstance(geom, Polygon):
                edges = self.__skeletonizePolygon(geom)
                if (0 < edges.length):
                    rows.append(self.updateOrAppend(row, { 'geometry': edges }))
            elif isinstance(geom, MultiPolygon):
                listOfEdges = []
                for _geom in geom.geoms:
                    edges = self.__skeletonizePolygon(geom)
                    if (0 < edges.length):
                        listOfEdges.append(edges)
                edges = unary_union(listOfEdges)
                if (0 < edges.length):
                    rows.append(self.updateOrAppend(row, { 'geometry': edges }))

        return GeoDataFrame(rows, crs=self.inputGdf.crs)



'''
Created on 14 oct. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from shapely.geometry import box, CAP_STYLE
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class STSquaredBBox(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, gdf, buffDist=0):
        '''
        Constructor
        '''
        if not isinstance(gdf, GeoDataFrame):
            raise IllegalArgumentTypeException(gdf, 'GeoDataFrame')
        self.gdf = gdf
        self.buffDist = buffDist

    def run(self):
        minx, miny, maxx, maxy = self.gdf.total_bounds
        buffDist = max(maxx - minx, maxy - miny) / 2.0 + self.buffDist
        bbox = box(minx, miny, maxx, maxy)
        bbox = bbox.centroid.buffer(buffDist, cap_style=CAP_STYLE.square)
        return GeoDataFrame([{'geometry': bbox}], crs=self.gdf.crs)



'''
Created on 23 mars 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import asarray, gradient, unique
from pandas import DataFrame
from scipy.cluster.hierarchy import dendrogram, linkage 
from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_score
from sklearn.mixture import GaussianMixture
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException

import matplotlib.pyplot as plt


class OptimalNumberOfClusters(GeoProcess):
    '''
    classdocs
    '''
    NUMERICS = ["int16", "int32", "int64", "float16", "float32", "float64"]

    def __init__(self, df, method="elbow", maxNumber=10, verbose=True):
        '''
        Constructor
        '''
        if not isinstance(df, DataFrame):
            raise IllegalArgumentTypeException(df, "DataFrame")
        self.df = df
        self.data = df.select_dtypes(include=self.NUMERICS)

        _fieldnames = list(self.data.columns)
        if (0 == len(_fieldnames)):
            raise Exception("There are no numeric fields in the DataFrame!")
        if verbose:
            print(f"The following fields are taken into account: {_fieldnames}")

        if not method in ("elbow", "silhouette", "bic", "dendrogram"):
            raise IllegalArgumentTypeException(method, "'elbow', 'silhouette', 'bic', or 'dendrogram'")
        self.method = method
        self.maxNumber = maxNumber
        self.verbose = verbose
        self.K = range(2, self.maxNumber + 1)
        self.sum_squared_dists = None
        self.silhouette_avg = None

    def _iterate_over_K(self):
        self.sum_squared_dists = []
        self.silhouette_avg = []

        for nclusters in self.K:
            kmeans = KMeans(n_clusters=nclusters, n_init="auto")
            kmeans.fit(self.data)

            self.sum_squared_dists.append(kmeans.inertia_)
            self.silhouette_avg.append(silhouette_score(self.data, kmeans.labels_))

        self.silhouette_avg = asarray(self.silhouette_avg)

    def _elbow(self):
        """
        Finding a "Kneedle" in a Haystack: Detecting Knee Points in System Behavior
        DOI: 10.1109/ICDCSW.2011.20
        https://raghavan.usc.edu//papers/kneedle-simplex11.pdf

        The point of maximum curvature is well-matched to the adhoc methods 
        operators use to select a knee, since curvature is a mathematical 
        measure of how much a function differs from a straight line. 
        """
        if self.sum_squared_dists is None:
            self._iterate_over_K()

        ssd = self.sum_squared_dists
        curvature = gradient(gradient(ssd)) / (1 + gradient(ssd) ** 2) ** 1.5
        nclusters = 1 + curvature.argmax()

        if self.verbose:
            plt.plot(self.K, ssd, "bx-")
            plt.xlabel("Nb of clusters") 
            plt.ylabel("Sum of squared distances / Inertia") 
            plt.title(f"Elbow Method For Optimal k ({nclusters})")
            plt.show()

        return nclusters

    def _silhouette(self):
        if self.silhouette_avg is None:
            self._iterate_over_K()
        nclusters = self.K[self.silhouette_avg.argmax()]

        if self.verbose:
            plt.plot(self.K, self.silhouette_avg, "bx-")
            plt.xlabel("Values of K") 
            plt.ylabel("Silhouette score") 
            plt.title(f"Silhouette analysis For Optimal k ({nclusters})")
            plt.show()

        return nclusters

    def _BICscore(self):
        """
        BAYESIAN INFORMATION CRITERION (BIC) SCORE
        The lower the BIC score, better is the model.
        We can use the BIC score for the Gaussian Mixture Modelling approach for clustering.
        """
        covariance_type = ["spherical", "tied", "diag", "full"]
        scores = []
        for cov in covariance_type:
            for nclusters in self.K:
                gmm = GaussianMixture(n_components=nclusters, covariance_type=cov)
                gmm.fit(self.data)
                scores.append((cov, nclusters, gmm.bic(self.data)))

        scores = DataFrame(scores, columns=["cov_type", "nclusters", "BIC"])
        idx = scores.groupby(by="cov_type").BIC.idxmin()
        scores = scores.loc[idx]
        if self.verbose:
            print(scores)

        return scores.loc[scores.BIC.idxmin(), "nclusters"]

    def _dendrogram(self):
        if self.verbose:
            plt.figure(figsize=(10, 7))  
            plt.title("Dendrogram")  
            dend = dendrogram(linkage(self.data, method='ward'))
            plt.show()
        else:
            dend = dendrogram(linkage(self.data, method='ward'))
        return len(unique(dend["leaves_color_list"]))

    def run(self):
        if ("elbow" == self.method):
            nclusters = self._elbow()
        elif ("silhouette" == self.method):
            nclusters = self._silhouette()
        elif ("bic" == self.method):
            nclusters = self._BICscore()
        elif ("dendrogram" == self.method):
            nclusters = self._dendrogram()

        return DataFrame(data=[{"method": self.method, "nclusters": nclusters}])



'''
Created on 31 mars 2021

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame, overlay, sjoin_nearest
from numpy import ceil, gradient, linspace, ndarray, sqrt, zeros
from shapely.geometry import box, LineString, MultiLineString
from t4gpd.commons.ArrayCoding import ArrayCoding
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.grid.AbstractGridLib import AbstractGridLib


class GridLib(AbstractGridLib):
    '''
    classdocs
    '''

    def __init__(self, gdf, dx, dy=None, encode=True):
        '''
        Constructor
        '''
        super().__init__(gdf, dx, dy, encode)

        self.minx, self.miny, maxx, maxy = gdf.total_bounds
        self.ncols = int(ceil((maxx - self.minx) / self.dx))
        self.nrows = int(ceil((maxy - self.miny) / self.dy))

        self.xOffset = ((self.ncols * self.dx) - (maxx - self.minx)) / 2.0
        self.yOffset = ((self.nrows * self.dy) - (maxy - self.miny)) / 2.0

    def __neighbors(self, r, c):
        e = r * self.ncols + (c + 1)
        ne = (r + 1) * self.ncols + (c + 1)
        n = (r + 1) * self.ncols + c
        nw = (r + 1) * self.ncols + (c - 1)
        w = r * self.ncols + (c - 1)
        sw = (r - 1) * self.ncols + (c - 1)
        s = (r - 1) * self.ncols + c
        se = (r - 1) * self.ncols + (c + 1)

        _na = -1
        if (0 == c):
            nw, w, sw = _na, _na, _na
        if (0 == r):
            sw, s, se = _na, _na, _na
        if (self.ncols == c + 1):
            ne, e, se = _na, _na, _na
        if (self.nrows == r + 1):
            nw, n, ne = _na, _na, _na
        return [e, n, w, s], [e, ne, n, nw, w, sw, s, se]

    def grid(self):
        result = []
        x0 = self.minx - self.xOffset
        for c in range(self.ncols):
            x0pp, y0 = x0 + self.dx, self.miny - self.yOffset
            for r in range(self.nrows):
                y0pp = y0 + self.dy
                neighbors4, neighbors8 = self.__neighbors(r, c)
                if self.encode:
                    neighbors4, neighbors8 = (ArrayCoding.encode(neighbors4),
                                              ArrayCoding.encode(neighbors8))
                result.append({
                    "gid": r * self.ncols + c,
                    "row": r,
                    "column": c,
                    "neighbors4": neighbors4,
                    "neighbors8": neighbors8,
                    "geometry": box(x0, y0, x0pp, y0pp)
                })
                y0 = y0pp
            x0 = x0pp

        return GeoDataFrame(result, crs=self.gdf.crs)

    @staticmethod
    def getGrid1(gdf, dx, dy=None):
        dy = dx if (dy is None) else dy

        minx, miny, maxx, maxy = gdf.total_bounds
        ncols = int(ceil((maxx - minx) / dx))
        nrows = int(ceil((maxy - miny) / dy))

        xOffset = ((ncols * dx) - (maxx - minx)) / 2.0
        yOffset = ((nrows * dy) - (maxy - miny)) / 2.0

        x0, y0 = minx - xOffset, miny - yOffset

        rows = []
        for c, x in enumerate(linspace(x0, x0 + ncols * dx, ncols, endpoint=False)):
            xpp = x+dx
            for r, y in enumerate(linspace(y0, y0 + nrows * dx, nrows, endpoint=False)):
                ypp = y+dy
                rows.append({
                    "gid": r * ncols + c,
                    "dx": dx,
                    "dy": dy,
                    "geometry": box(x, y, xpp, ypp)
                })
        grid = GeoDataFrame(rows, crs=gdf.crs)
        grid2 = GridLib.distanceToNearestContour(gdf, grid)
        return grid2

    @staticmethod
    def __neighbors_4_8(nrows, ncols, r, c):
        e = r * ncols + (c + 1)
        ne = (r + 1) * ncols + (c + 1)
        n = (r + 1) * ncols + c
        nw = (r + 1) * ncols + (c - 1)
        w = r * ncols + (c - 1)
        sw = (r - 1) * ncols + (c - 1)
        s = (r - 1) * ncols + c
        se = (r - 1) * ncols + (c + 1)

        _na = -1
        if (0 == c):
            nw, w, sw = _na, _na, _na
        if (0 == r):
            sw, s, se = _na, _na, _na
        if (ncols == c + 1):
            ne, e, se = _na, _na, _na
        if (nrows == r + 1):
            nw, n, ne = _na, _na, _na
        return [e, n, w, s], [e, ne, n, nw, w, sw, s, se]

    @staticmethod
    def getGrid2(gdf, dx, dy=None, encode=True):
        dy = dx if (dy is None) else dy

        minx, miny, maxx, maxy = gdf.total_bounds
        ncols = int(ceil((maxx - minx) / dx))
        nrows = int(ceil((maxy - miny) / dy))

        xOffset = ((ncols * dx) - (maxx - minx)) / 2.0
        yOffset = ((nrows * dy) - (maxy - miny)) / 2.0

        x0, y0 = minx - xOffset, miny - yOffset

        rows = []
        for c, x in enumerate(linspace(x0, x0 + ncols * dx, ncols, endpoint=False)):
            xpp = x+dx
            for r, y in enumerate(linspace(y0, y0 + nrows * dx, nrows, endpoint=False)):
                ypp = y+dy
                neighbors4, neighbors8 = GridLib.__neighbors_4_8(
                    nrows, ncols, r, c)
                if encode:
                    neighbors4, neighbors8 = (ArrayCoding.encode(neighbors4),
                                              ArrayCoding.encode(neighbors8))
                rows.append({
                    "gid": r * ncols + c,
                    "row": r,
                    "column": c,
                    "neighbors4": neighbors4,
                    "neighbors8": neighbors8,
                    # "dx": dx,
                    # "dy": dy,
                    "geometry": box(x, y, xpp, ypp)
                })
        grid = GeoDataFrame(rows, crs=gdf.crs)
        grid2 = GridLib.distanceToNearestContour(gdf, grid)
        return grid2

    @staticmethod
    def distanceToNearestContour(gdf, grid):
        gdf = gdf.copy(deep=True)
        gdf.geometry = gdf.geometry.apply(
            lambda geom: MultiLineString(GeomLib.toListOfLineStrings(geom)))
        grid2 = sjoin_nearest(
            grid, gdf[["geometry"]], how="inner", distance_col="dist_to_ctr", max_distance=None)
        if ("index_right" not in gdf) and ("index_right" in grid2):
            grid2.drop(columns="index_right", inplace=True)
        grid2.drop_duplicates(
            subset=["gid"], keep="first", ignore_index=True, inplace=True)
        return grid2

    @staticmethod
    def fromGridToNumpyArray(gdf, fieldvalue, rowFieldname="row", colFieldname="column"):
        if not isinstance(gdf, GeoDataFrame):
            raise IllegalArgumentTypeException(gdf, "GeoDataFrame")
        for fieldname in [fieldvalue, rowFieldname, colFieldname]:
            if not fieldname in gdf:
                raise Exception(f"{fieldname} is not a relevant field name!")

        nrows = 1 + gdf[rowFieldname].max()
        ncols = 1 + gdf[colFieldname].max()
        result = zeros((nrows, ncols))

        for _, row in gdf.iterrows():
            r, c, v = row[rowFieldname], row[colFieldname], row[fieldvalue]
            if not ((r is None) or (c is None) or (v is None)):
                result[r, c] = v
        return result

    @staticmethod
    def fromNumpyArrayToGrid(nparray, gdf, fieldvalue, rowFieldname="row", colFieldname="column"):
        if not isinstance(nparray, ndarray):
            raise IllegalArgumentTypeException(gdf, "NumPy Array")
        if not isinstance(gdf, GeoDataFrame):
            raise IllegalArgumentTypeException(gdf, "GeoDataFrame")
        for fieldname in [rowFieldname, colFieldname]:
            if not fieldname in gdf:
                raise Exception(f"{fieldname} is not a relevant field name!")

        result = gdf.copy(deep=True)
        result[fieldvalue] = 0
        for idx, row in result.iterrows():
            r, c = row[rowFieldname], row[colFieldname]
            result.at[idx, fieldvalue] = nparray[r, c]
        return result

    @staticmethod
    def gradient(gdf, fieldvalue, rowFieldname="row", colFieldname="column", magn=1.0, normalize=False):
        nparray = GridLib.fromGridToNumpyArray(
            gdf, fieldvalue, rowFieldname, colFieldname)
        # nparray = nparray.T
        UU, VV = gradient(nparray)

        result = gdf.copy(deep=True)
        result["grad_x"], result["grad_y"] = 0, 0
        for idx, row in result.iterrows():
            r, c = row[rowFieldname], row[colFieldname]
            g = row.geometry.centroid.coords[0][0:2]
            grad_x, grad_y = UU[r, c], VV[r, c]

            if normalize:
                _magn = magn / sqrt(grad_x ** 2 + grad_y ** 2)
            else:
                _magn = magn
            grad_x, grad_y = _magn * grad_x, _magn * grad_y

            result.at[idx, "grad_x"] = grad_x
            result.at[idx, "grad_y"] = grad_y
            result.at[idx, "geometry"] = LineString(
                [g, [g[0] + grad_x, g[1] + grad_y]])
        return result

    @staticmethod
    def divergence(gdf, fieldvalue, rowFieldname="row", colFieldname="column", magn=1.0):
        nparray = GridLib.fromGridToNumpyArray(
            gdf, fieldvalue, rowFieldname, colFieldname)
        # nparray = nparray.T
        UU, VV = gradient(nparray)

        result = gdf.copy(deep=True)
        result["divergence"] = 0
        for idx, row in result.iterrows():
            r, c = row[rowFieldname], row[colFieldname]
            grad_x, grad_y = UU[r, c], VV[r, c]

            result.at[idx, "divergence"] = grad_x + grad_y
        return result

    @staticmethod
    def __overlayWithGrid(gdf, grid, how="difference"):
        _result = grid.loc[:, ["gid", "geometry"]]
        _result.geometry = _result.geometry.apply(lambda geom: geom.centroid)
        _result = overlay(_result, gdf, how)
        _result = set(_result.gid.to_list())
        return _result

    @staticmethod
    def getIndoorOutdoorGrid(gdf, grid):
        grid2 = grid.copy(deep=True)
        gids = GridLib.__overlayWithGrid(gdf, grid2, how="difference")
        grid2["indoor"] = grid2.gid.apply(
            lambda gid: 0 if (gid in gids) else 1)
        return grid2

    @staticmethod
    def getIndoorGrid(gdf, grid):
        grid2 = GridLib.getIndoorOutdoorGrid(gdf, grid)
        grid2 = grid2.loc[grid2[grid2.indoor == 1].index]
        grid2.reset_index(drop=True, inplace=True)
        return grid2

    @staticmethod
    def getOutdoorGrid(gdf, grid):
        grid2 = GridLib.getIndoorOutdoorGrid(gdf, grid)
        grid2 = grid2.loc[grid2[grid2.indoor == 0].index]
        grid2.reset_index(drop=True, inplace=True)
        return grid2

    @staticmethod
    def getSubgrid1(gdf, grid, dx, threshold):
        gid = grid.gid.max()
        rows = []
        for _, row in grid.iterrows():
            if (row.dist_to_ctr > threshold):
                rows.append({
                            "gid": row.gid,
                            "dx": row.dx,
                            "geometry": row.geometry
                            })
            else:
                x0, y0, x1, y1 = row.geometry.bounds
                ncols = int(round((x1 - x0) / dx))
                nrows = int(round((y1 - y0) / dx))
                for x in linspace(x0, x1, ncols, endpoint=False):
                    xpp = x+dx
                    for y in linspace(y0, y1, nrows, endpoint=False):
                        ypp = y+dx
                        gid += 1
                        rows.append({
                            "gid": gid,
                            "dx": dx,
                            "geometry": box(x, y, xpp, ypp)
                        })
        grid2 = GeoDataFrame(rows, crs=gdf.crs)
        grid3 = GridLib.distanceToNearestContour(gdf, grid2)
        return grid3



'''
Created on 15 sept. 2022

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from glob import glob
from os.path import basename, exists
from sys import stderr
import warnings

from geopandas import GeoDataFrame
from pandas import concat
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.WarnUtils import WarnUtils
from t4gpd.picoclim.CampbellSciReader import CampbellSciReader
from t4gpd.picoclim.ExtraProcessing import ExtraProcessing
from t4gpd.picoclim.InertialMeasureReader import InertialMeasureReader
from t4gpd.picoclim.JoinByTimeDistance import JoinByTimeDistance
from t4gpd.picoclim.KestrelReader import KestrelReader
from t4gpd.picoclim.MeteoFranceReader import MeteoFranceReader
from t4gpd.picoclim.SensirionReader import SensirionReader
from t4gpd.picoclim.SnapImuOnTrackUsingWaypoints import SnapImuOnTrackUsingWaypoints
from t4gpd.picoclim.SnapUclimOnTrackUsingWaypoints import SnapUclimOnTrackUsingWaypoints
from t4gpd.picoclim.TracksWaypointsReader import TracksWaypointsReader
from t4gpd.picoclim.UClimGuidingReader import UClimGuidingReader
from t4gpd.picoclim.UClimTrackWaypointsReader import UClimTrackWaypointsReader


class MetrologicalCampaignReader(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, dirName, dirMeteoFr=None, dirUclimTrack=None):
        '''
        Constructor
        '''
        warnings.formatwarning = WarnUtils.format_Warning_alt

        if not exists(dirName):
            raise IllegalArgumentTypeException(dirName, "valid directory name")
        self.dirName = dirName

        if not ((dirMeteoFr is None) or exists(dirMeteoFr)):
            raise IllegalArgumentTypeException(
                dirMeteoFr, "valid directory name")
        self.dirMeteoFr = dirMeteoFr

        if not ((dirUclimTrack is None) or exists(dirUclimTrack)):
            raise IllegalArgumentTypeException(
                dirUclimTrack, "valid directory name")
        self.dirUclimTrack = dirUclimTrack

        self.performance = None, None, None, None, None, None

    def getVersion(self):
        uclimLOC = glob(f"{self.dirName}/uclimLOC_*_[0-9]*T[0-9]*.txt")
        if (0 == len(uclimLOC)):
            # OLD LOCATION SYSTEM BASED ON INERTIAL SENSOR
            return 1
        elif (1 == len(uclimLOC)):
            # NEW LOCATION SYSTEM BASED ON UCLIM GUIDING SOLUTION
            return 2
        raise NotImplementedError(
            f'\n\n\t*** MetrologicalCampaignReader:: unknown format version! ***')

    def getPerformance(self):
        return self.performance

    def __loader(self, filenameFilter, Reader):
        dfs = []
        for filename in glob(filenameFilter):
            print(f"*** {basename(filename)}", file=stderr)
            df = Reader(filename).run()
            dfs.append(df)
        return None if (0 == len(dfs)) else concat(dfs, ignore_index=True)

    def __readSig(self):
        filename = f"{self.dirName}/sig.gpkg"
        print(f"*** {basename(filename)}", file=stderr)
        dfStatic, tracks, waypoints = TracksWaypointsReader(filename).run()
        return dfStatic, tracks, waypoints

    def __readUclimGuiding(self):
        filenameFilter = f"{self.dirName}/uclimLOC_*_[0-9]*T[0-9]*.txt"
        # dfUclim = self.__loader(filenameFilter, UClimGuidingReader)
        reader = UClimGuidingReader(glob(filenameFilter)[0])
        dfUclim = reader.run()
        dt0, dt1, deltaTotal, deltaActual, performance = reader.getPerformance()

        trackid = dfUclim.track_id.unique().astype(int)
        if (1 != len(trackid)):
            raise Exception(
                "There must be exactly one and only one track in the uclimLOC file!")
        trackid = trackid[0]

        if self.dirUclimTrack is None:
            raise Exception("dirUclimTrack must not be None!")

        uclimTrack = glob(f"{self.dirUclimTrack}/*_track{trackid}.csv")
        if (1 == len(uclimTrack)):
            tracks, waypoints = UClimTrackWaypointsReader(uclimTrack[0]).run()
        else:
            raise Exception(
                f"There's no file: {self.dirUclimTrack}/*_track{trackid}.csv")

        self.performance = dt0, dt1, deltaTotal, deltaActual, \
            tracks.geometry.squeeze().length / deltaActual, performance
        return dfUclim, tracks, waypoints

    def __readImu(self):
        filenameFilter = f"{self.dirName}/LOG_FILE_????????_*.txt"
        dfImu = self.__loader(filenameFilter, InertialMeasureReader)
        if dfImu is not None:
            dfImu.sort_values(by="timestamp", inplace=True, ignore_index=True)
        return dfImu

    def __readMob(self):
        filenameFilter = f"{self.dirName}/CR1000XSeries_TwoSec.dat"
        dfMob = self.__loader(filenameFilter, CampbellSciReader)
        if dfMob is not None:
            dfMob.sort_values(by="timestamp", inplace=True, ignore_index=True)
        return dfMob

    def __readOther(self, dfStatic=None):
        filenameFilter = f"{self.dirName}/*HEAT*.csv"
        dfStat1 = self.__loader(filenameFilter, KestrelReader)
        if dfStat1 is not None:
            dfStat1.sort_values(
                by="timestamp", inplace=True, ignore_index=True)

        filenameFilter = f"{self.dirName}/Sensirion_*.edf"
        dfStat2 = self.__loader(filenameFilter, SensirionReader)
        if dfStat2 is not None:
            dfStat2.sort_values(
                by="timestamp", inplace=True, ignore_index=True)
            if dfStatic is not None:
                dfStat2 = dfStat2.merge(dfStatic, on="station")
                dfStat2 = GeoDataFrame(dfStat2, crs=dfStatic.crs)

        filenameFilter = f"{self.dirMeteoFr}/meteo-france-*.txt"
        dfMeteoFr = self.__loader(filenameFilter, MeteoFranceReader)
        if dfMeteoFr is not None:
            dfMeteoFr.sort_values(
                by="timestamp", inplace=True, ignore_index=True)

        return dfStat1, dfStat2, dfMeteoFr

    def __readV1(self):
        warnings.warn("MetrologicalCampaignReader (v1)")

        dfImu = self.__readImu()
        dfMob = self.__readMob()
        dfStatic, tracks, waypoints = self.__readSig()

        if dfImu is None:
            raise Exception("dfImu must not be None!")
        dfImu = dfImu.to_crs(tracks.crs)
        dfImu = SnapImuOnTrackUsingWaypoints(dfImu, tracks, waypoints).run()
        dfImuMob = JoinByTimeDistance(
            dfImu, dfMob, left_on="timestamp", right_on="timestamp").run()
        dfImuMob = ExtraProcessing(dfImuMob, inplace=False).run()

        dfStat1, dfStat2, dfMeteoFr = self.__readOther(dfStatic)

        return dfStatic, tracks, waypoints, dfImuMob, dfMob, dfStat1, dfStat2, dfMeteoFr

    def __readV2(self):
        warnings.warn("MetrologicalCampaignReader (v2)")

        dfMob = self.__readMob()

        dfUclim, tracks, waypoints = self.__readUclimGuiding()
        dfUclim = SnapUclimOnTrackUsingWaypoints(
            dfUclim, dfMob, tracks, waypoints).run()

        dfUclimMob = dfUclim.merge(dfMob, how="inner", on="timestamp")
        dfUclimMob = ExtraProcessing(dfUclimMob, inplace=False).run()

        dfStat1, dfStat2, dfMeteoFr = self.__readOther()

        return None, tracks, waypoints, dfUclimMob, dfMob, dfStat1, dfStat2, dfMeteoFr

    def run(self):
        version = self.getVersion()

        if (1 == version):
            return self.__readV1()
        elif (2 == version):
            return self.__readV2()

        raise Exception("Unreachable instruction!")


'''
version = 2
if (1 == version):
    dirName1 = "/home/tleduc/prj/nm-ilots-frais/terrain/220711"
    # dirName1 = "/home/tleduc/prj/nm-ilots-frais/terrain/220713"
    dirName2 = "/home/tleduc/prj/nm-ilots-frais/information"

    dfStatic, tracks, waypoints, dfImuMob, dfMob, dfStat1, dfStat2, dfMeteoFr = \
        MetrologicalCampaignReader(dirName1, dirMeteoFr=dirName2).run()
else:
    dirName1 = "/home/tleduc/prj/uclim/data/nantes_commerce_feydeau/nantes_commerce_feydeau_20230607_track1_093626/raw_data"
    dirName2 = None
    dirName3 = "/home/tleduc/prj/uclim/data/nantes_commerce_feydeau"

    mcr = MetrologicalCampaignReader(dirName1, dirMeteoFr=dirName2, dirUclimTrack=dirName3)
    _, tracks, waypoints, dfUclimMob, dfMob, dfStat1, dfStat2, dfMeteoFr = mcr.run()
    print(mcr.getPerformance())
'''



'''
Created on 11 mai 2021

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import isnan
from pandas import DataFrame, Interval, IntervalIndex
from t4gpd.comfort.algo.UTCILib import UTCILib
from t4gpd.comfort.indices.AbstractThermalComfortIndice import AbstractThermalComfortIndice
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class UTCI(AbstractThermalComfortIndice):
    '''
    classdocs
    '''
    RANGES = DataFrame(
        data=[("Extreme cold stress", "#003075"),
              ("Very strong cold stress", "#0088cf"),
              ("Strong cold stress", "#00c2f7"),
              ("Moderate cold stress", "#91c1e1"),
              ("Slight cold stress", "#dbebf5"),
              ("No thermal stress", "#ffffff"),
              ("Moderate heat stress", "#fffa00"),
              ("Strong heat stress", "#ff7900"),
              ("Very strong heat stress", "#f40000"),
              ("Extreme heat stress", "#ff0000")],
        columns=["label", "color"],
        index=IntervalIndex([
            Interval(left=-float("inf"), right=-40, closed="right"),
            Interval(left=-40, right=-27, closed="right"),
            Interval(left=-27, right=-13, closed="right"),
            Interval(left=-13, right=0, closed="right"),
            Interval(left=0, right=9, closed="right"),
            Interval(left=9, right=26, closed="right"),
            Interval(left=26, right=32, closed="right"),
            Interval(left=32, right=38, closed="right"),
            Interval(left=38, right=46, closed="right"),
            Interval(left=46, right=100, closed="right"),
            # Interval(left=46, right=float("inf"), closed="right"),
            ])
        )

    def __init__(self, sensorsGdf, AirTC="AirTC_Avg", RH="RH_Avg", WS_ms="WS_ms_Avg",
                 T_mrt="T_mrt"):
        '''
        Constructor

        AirTC: air temperature [C]
        RH: relative humidity [%]
        Ws_ms: wind speed recorded at pedestrian level (at height 1.1 m) [m.s-1]
        T_mrt: Mean radiant temperature [C]
        '''
        if not isinstance(sensorsGdf, DataFrame):
            raise IllegalArgumentTypeException(sensorsGdf, "DataFrame")

        for fieldname in (AirTC, RH, WS_ms, T_mrt):
            if fieldname not in sensorsGdf:
                raise Exception("%s is not a relevant field name!" % fieldname)

        self.AirTC = AirTC
        self.RH = RH
        self.WS_ms = WS_ms
        self.T_mrt = T_mrt

    @staticmethod
    def thermalPerceptionRange(tp):
        return UTCI.RANGES.iloc[UTCI.RANGES.index.get_loc(tp)].label

    @staticmethod
    def thermalPerceptionRanges():
        # Excerpt from https://doi.org/10.1016/j.wace.2018.01.004
        return {
            row.label: {"min": idx.left, "max": idx.right, "color": row.color} for idx, row in UTCI.RANGES.iterrows()
            }

    def runWithArgs(self, row):
        AirTC = row[self.AirTC]
        RH = row[self.RH]
        WS_ms = row[self.WS_ms]
        T_mrt = row[self.T_mrt]

        UTCI = None
        if not (isnan(AirTC) or isnan(RH) or isnan(WS_ms) or isnan(T_mrt)):
            # Universal Thermal Climate Index (UTCI)
            UTCI = UTCILib.assess_utci(AirTC, RH, WS_ms, T_mrt)

        return { "UTCI": UTCI }



'''
Created on 31 janv. 2023

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from numpy import asarray, sqrt
from shapely.geometry import Polygon
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class STVariableWidthBuffer(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, gdf, widthFieldname):
        '''
        Constructor
        '''
        if not isinstance(gdf, GeoDataFrame):
            raise IllegalArgumentTypeException(gdf, 'GeoDataFrame')
        self.gdf = gdf
        self.widthFieldname = widthFieldname

    def __dilate(self, p0, p1, p2, width):
        if (p0 is None):
            before, after = p1, p2
        elif (p2 is None):
            before, after = p0, p1
        else:
            before, after = p0, p2
        # mv: Motion vector; nmv: L2-norm of the motion vector; nv: Normal vector
        mv = asarray([after.x - before.x, after.y - before.y])
        nmv = sqrt((mv ** 2).sum())
        if (0 == nmv):
            return [p1.x, p1.y], [p1.x, p1.y]
        mv = mv / nmv
        nv = [mv[1], -mv[0]]
        return (
            [p1.x + width * nv[0], p1.y + width * nv[1]],
            [p1.x - width * nv[0], p1.y - width * nv[1]]
            )
        
    def run(self):
        pairs = []
        for _, row in self.gdf.iterrows():
            pairs.append({'geometry': row.geometry, 'width': row[self.widthFieldname]})
        n = len(pairs)

        rows1, rows2 = [], []
        for i in range(n):
            p0 = pairs[i - 1]['geometry'] if (0 < i) else None
            p1, width = pairs[i]['geometry'], pairs[i]['width']
            p2 = pairs[i + 1]['geometry'] if (i + 1 < n) else None
            down, up = self.__dilate(p0, p1, p2, width)
            rows1.append(down)
            rows2.insert(0, up)

        geom = Polygon(rows1 + rows2)
        return GeoDataFrame([{'geometry': geom}], crs=self.gdf.crs)



'''
Created on 4 mars 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import datetime, timezone

from numpy import float64, ndarray
from t4gpd.commons.ArrayCoding import ArrayCoding
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess

from t4gpd.avidon.commons.EnergyDemandCalculator2 import EnergyDemandCalculator2


class EnergyDemandOfITEquipment3(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, columns, dt, tz=timezone.utc, D=None, S=None, E=None, encode=False):
        '''
        Constructor
        '''
        self.columns = columns
        self.encode = encode

        if isinstance(dt, datetime):
            dt = dt.replace(tzinfo=tz) if dt.tzinfo is None else dt
            self.dt = [ dt ]
            self.calc = [ EnergyDemandCalculator2(dt, D, S, E) ]
        elif isinstance(dt, (list, ndarray, tuple)):
            self.dt, self.calc = [], []
            for _dt in dt:
                if not isinstance(_dt, datetime):
                    raise IllegalArgumentTypeException(dt, 'datetime or list of datetimes')
                _dt = _dt.replace(tzinfo=tz) if _dt.tzinfo is None else _dt
                self.dt.append(_dt)
                self.calc.append(EnergyDemandCalculator2(_dt, D, S, E))
        else:
            raise IllegalArgumentTypeException(dt, 'datetime or list of datetimes')

    def runWithArgs(self, row):
        _C = row[self.columns].to_numpy(dtype=float64)

        whs = [_calc.energyDemandOfITEquipment(_C)  for _calc in self.calc]

        if (1 == len(self.dt)):
            return { 'IT_in_Wh': whs[0], 'datetime': self.dt[0] }

        if self.encode:
            return { 'IT_in_Wh': ArrayCoding.encode(whs), 'datetime': ArrayCoding.encode(self.dt) }

        return { 'IT_in_Wh': whs, 'datetime': self.dt }



'''
Created on 17 fev. 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from shapely.geometry import box, JOIN_STYLE
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class STBBox(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, gdf, buffDist=0):
        '''
        Constructor
        '''
        if not isinstance(gdf, GeoDataFrame):
            raise IllegalArgumentTypeException(gdf, 'GeoDataFrame')
        self.gdf = gdf
        self.buffDist = buffDist

    def run(self):
        bbox = box(*self.gdf.total_bounds).buffer(self.buffDist, join_style=JOIN_STYLE.mitre)
        return GeoDataFrame([{'geometry': bbox}], crs=self.gdf.crs)



'''
Created on 27 mai 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import datetime, timezone

from t4gpd.avidon.scenarios.AbstractScenario import AbstractScenario
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess

class EnergyDemandOfITEquipment(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, scenario, dt, tz=timezone.utc):
        '''
        Constructor
        '''
        if not isinstance(scenario, AbstractScenario):
            raise IllegalArgumentTypeException(scenario, 'AbstractScenario')
        self.scenario = scenario

        if not isinstance(dt, datetime):
            raise IllegalArgumentTypeException(dt, 'datetime')
        self.dt = dt.replace(tzinfo=tz) if dt.tzinfo is None else dt

    def runWithArgs(self, row):
        return {
            'IT_in_Wh': self.scenario.energyDemandOfITEquipment(row, self.dt),
            'datetime': self.dt
            }



'''
Created on 3 oct 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import matplotlib.pyplot as plt
from datetime import datetime
from numpy import linspace, pi
from t4gpd.energy.Angstrom import Angstrom
from t4gpd.energy.Dogniaux import Dogniaux
from t4gpd.energy.Perez import Perez
from t4gpd.energy.Perraudeau import Perraudeau
from t4gpd.energy.PerrinDeBrichambaut import PerrinDeBrichambaut


class PlotIrradiances(object):
    '''
    classdocs
    '''
    @staticmethod
    def plot1(dt=datetime(2021, 6, 21), ofile=None):
        # PEREZ MODEL PARAMETERS:
        delta = 0.12
        epsilon = 6.3
        dayInYear = dt.timetuple().tm_yday

        Xdeg, Xrad = linspace(0, 90, 91), linspace(0, pi/2, 91)

        y1 = [Dogniaux.directNormalIrradiance(x) for x in Xrad]
        y2 = [Perez.directSolarIrradiance(
            x, dayInYear, delta, epsilon) for x in Xrad]
        y3 = [Perraudeau.directNormalIrradiance(x) for x in Xrad]
        y4 = [PerrinDeBrichambaut.directNormalIrradiance(
            x, skyType=PerrinDeBrichambaut.PURE_SKY) for x in Xrad]
        y5 = [PerrinDeBrichambaut.directNormalIrradiance(
            x, skyType=PerrinDeBrichambaut.STANDARD_SKY) for x in Xrad]
        y6 = [PerrinDeBrichambaut.directNormalIrradiance(
            x, skyType=PerrinDeBrichambaut.POLLUTED_SKY) for x in Xrad]
        y7 = [Perez.diffuseSolarIrradiance(x, dayInYear, delta) for x in Xrad]
        y8 = [PerrinDeBrichambaut.diffuseSolarIrradiance(
            x, skyType=PerrinDeBrichambaut.PURE_SKY) for x in Xrad]
        y9 = [PerrinDeBrichambaut.diffuseSolarIrradiance(
            x, skyType=PerrinDeBrichambaut.STANDARD_SKY) for x in Xrad]
        ya = [PerrinDeBrichambaut.diffuseSolarIrradiance(
            x, skyType=PerrinDeBrichambaut.POLLUTED_SKY) for x in Xrad]

        fig, ax = plt.subplots(figsize=(1.5 * 8.26, 1.5 * 8.26))
        ax.set_title(
            "Direct Normal Irradiance (DNI) + Diffuse Horizontal Irradiance (DHI)", fontsize=16)
        ax.set_xlabel("Solar Altitude Angle [$^o$]", fontsize=16)
        ax.set_ylabel("Irradiance [W.m$^{-2}$]", fontsize=16)
        ax.set_ylim(0, 1200)
        ax.grid(True)
        _lines = ax.plot(Xdeg, y1, "b-", Xdeg, y2, "r-", Xdeg, y3, "k-", Xdeg, y4, "m-.",
                         Xdeg, y5, "g-.", Xdeg, y6, "c-.", Xdeg, y7, "r--", Xdeg, y8, "m--",
                         Xdeg, y9, "g--", Xdeg, ya, "c--")
        ax.legend(_lines, ["Dogniaux's DNI", "Perez's DNI", "Perraudeau's DNI",
                           "Perrin de Br.'s DNI (Pure sky)",
                           "Perrin de Br.'s DNI (Standard sky)",
                           "Perrin de Br.'s DNI (Polluted sky)",
                           "Perez's diffuse model",
                           "Perrin de Br.'s diffuse model (Pure sky)",
                           "Perrin de Br.'s diffuse model (Standard sky)",
                           "Perrin de Br.'s diffuse model (Polluted sky)"
                           ], loc="upper center", ncol=2, fontsize=12)
        if ofile is None:
            plt.show()
        else:
            plt.savefig(ofile, bbox_inches="tight")
        plt.close(fig)

    @staticmethod
    def plot2(ofile=None):
        Xdeg, Xrad = linspace(0, 90, 91), linspace(0, pi/2, 91)
        Y1 = [PerrinDeBrichambaut.directNormalIrradiance(
            x, skyType=PerrinDeBrichambaut.STANDARD_SKY) for x in Xrad]
        Y2 = [PerrinDeBrichambaut.diffuseSolarIrradiance(
            x, skyType=PerrinDeBrichambaut.STANDARD_SKY) for x in Xrad]

        fig, ax = plt.subplots(figsize=(1.5 * 8.26, 1.5 * 8.26))
        ax.set_ylim(0, 1200)
        ax.set_xlabel("Solar altitude angle [$^o$]", fontsize=20)
        ax.set_ylabel("Irradiance [W.m$^{-2}$]", fontsize=20)
        ax.grid(True)
        ax.plot(Xdeg, Y1, label="Direct Normal Irradiance (DNI)",
                linestyle="dotted", linewidth=3, color="black")
        ax.plot(Xdeg, Y2, label="Diffuse Horizontal Irradiance (DHI)",
                linestyle="dashdot", linewidth=3, color="black")

        nuples = [
            (20, 50, 8, "red", "solid"),
            (20, 50, 4, "green", "solid"),
            (20, 50, 0, "blue", "solid"),
            (30, 75, 8, "red", "dashed"),
            (30, 75, 4, "green", "dashed"),
            (30, 75, 0, "blue", "dashed"),
        ]
        for Tair, RH, N, color, ls in nuples:
            Y3 = [Angstrom.lowg_wave_irradiance(Tair, RH, N) for _ in Xdeg]
            ax.plot(
                Xdeg, Y3, label=f"LW from the atmos. (Ta={Tair:.0f}, RH={RH}%, N={N})", linestyle=ls, color=color)
        ax.legend()

        if ofile is None:
            plt.show()
        else:
            plt.savefig(ofile, bbox_inches="tight")
        plt.close(fig)


# PlotIrradiances.plot1()
# PlotIrradiances.plot2()



'''
Created on 15 sept. 2022

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from copy import deepcopy
from os.path import basename
import warnings
from zoneinfo import ZoneInfo

from dateutil.parser import parse
from geopandas import GeoDataFrame
from numpy.random import choice
from pandas import read_csv
from shapely.geometry import Point
from t4gpd.commons.WarnUtils import WarnUtils
from t4gpd.io.AbstractReader import AbstractReader


class InertialMeasureReader(AbstractReader):
    '''
    classdocs

    https://inertialelements.com/osmium-mimu22bl.html
    '''
    FIELDS = {
        "timestamp": str, "step_count": int, "X": float, "Y": float, "Z": float,
        "Distance": float, "degree": float, "latitude": float, "longitude": float,
        "GpsAccuracy": float, "indoor_outdoor_flag": str, "TagName": float
    }

    def __init__(self, inputFile, outputCrs=None, tzinfo="Europe/Paris"):
        '''
        Constructor
        '''
        warnings.formatwarning = WarnUtils.format_Warning
        self.inputFile = inputFile
        self.outputCrs = outputCrs
        self.tzinfo = ZoneInfo(tzinfo)

    @staticmethod
    def getVersion(inputFile):
        VERSIONS = [
            "     time_stamp",
            "\"TIME.STAMP\"",
            " TIME STAMP",
            "     TIME STAMP",
            "timestamp"            
            ]
        with InertialMeasureReader.opener(inputFile) as f:
            line = next(f)
            for i, version in enumerate(VERSIONS, start=0):
                if line.startswith(version):
                    return i
            raise NotImplementedError(f'\n\n\t*** InertialMeasureReader:: unknown file format version! ***')

    def __extract_YYYYMMDD_from_filename(self):
        yyyymmdd = basename(self.inputFile)[9:17]
        year, month, day = yyyymmdd[0:4], yyyymmdd[4:6], yyyymmdd[6:8]
        return year, month, day

    def __getRndSubtrackId(self):
        return "".join(choice([str(i) for i in range(10)], size=10))

    def __readV01(self):
        df = read_csv(self.inputFile, decimal=",", header=None, sep="\s+",
                      names=list(self.FIELDS.keys()),
                      skip_blank_lines=True, skiprows=2)
        for fieldname, fieldtype in self.FIELDS.items():
            if str(fieldtype) == "<class 'int'>":
                df[fieldname] = df[fieldname].astype(int)
            elif str(fieldtype) == "<class 'float'>":
                df[fieldname] = df[fieldname].astype(float)
        return df

    def __readV23(self):
        df = read_csv(self.inputFile, decimal=",", header=None, sep="\s+",
                      names=list(self.FIELDS.keys()), dtype=self.FIELDS,
                      skip_blank_lines=True, skiprows=2)
        return df

    def __readV4(self):
        FIELDS = deepcopy(self.FIELDS)
        FIELDS["Distance"] = str

        df = read_csv(self.inputFile, decimal=",", header=None, sep="\s+",
                      names=list(self.FIELDS.keys()), dtype=FIELDS,
                      skip_blank_lines=True, skiprows=1)
        return df

    def run(self):
        version = self.getVersion(self.inputFile)
        # warnings.warn(f"InertialMeasureReader ({version})")
        if (version in [0, 1]):
            df = self.__readV01()
        elif (version in [2, 3]):
            df = self.__readV23()
        elif (4 == version):
            df = self.__readV4()

        year, month, day = self.__extract_YYYYMMDD_from_filename()
        df.timestamp = df.timestamp.apply(lambda t: parse(f"{year}-{month}-{day}T{t}"))

        df.timestamp = df.timestamp.apply(lambda dt: dt.replace(tzinfo=self.tzinfo))
        if not df.timestamp.is_monotonic_increasing:
            msg = f"Timestamps in {self.inputFile} are not monotonically increasing"
            warnings.warn(msg)

        df["subtrack"] = self.__getRndSubtrackId()

        df["geometry"] = df.apply(lambda row: Point(row.longitude, row.latitude), axis=1)
        df = GeoDataFrame(df, geometry="geometry", crs="epsg:4326")

        if self.outputCrs is None:
            return df
        return df.to_crs(self.outputCrs)

"""
inputFile = "/home/tleduc/prj/nm-ilots-frais/terrain/220711/LOG_FILE_20220711_172000.txt"
df = InertialMeasureReader(inputFile, outputCrs="epsg:2154").run()
print(df.head(3))

inputFile = "/home/tleduc/prj/coolscapes-tl/dev/data/05-Quai-de-plantes-Nantes/20200526/localisation/LOG_FILE_20200526_120307.txt"
df = InertialMeasureReader(inputFile, outputCrs="epsg:2154").run()
print(df.head(3))

inputFile = "/home/tleduc/prj/coolscapes-tl/dev/data/06-Carriere-misery-Nantes/20200722/localisation/LOG_FILE_20200722_141610.txt"
df = InertialMeasureReader(inputFile, outputCrs="epsg:2154").run()
print(df.head(3))

inputFile = "/home/tleduc/prj/coolscapes-tl/dev/data/08-La-defense-Paris/20210721/localisation/LOG_FILE_20210721_182401.txt"
df = InertialMeasureReader(inputFile, outputCrs="epsg:2154").run()
print(df.head(3))
"""
